# SPDX-License-Identifier: MIT

from __future__ import annotations

import asyncio
import datetime
import itertools
import sys
from operator import attrgetter
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    List,
    Literal,
    Optional,
    Sequence,
    Tuple,
    Union,
    cast,
    overload,
)

import disnake.abc

from . import utils
from .activity import ActivityTypes, create_activity
from .asset import Asset
from .colour import Colour
from .enums import Status, try_enum
from .flags import MemberFlags
from .object import Object
from .permissions import Permissions
from .user import BaseUser, User, _UserTag
from .utils import MISSING

__all__ = (
    "VoiceState",
    "Member",
)

if TYPE_CHECKING:
    from typing_extensions import Self

    from .abc import Snowflake
    from .channel import DMChannel, StageChannel, VoiceChannel
    from .flags import PublicUserFlags
    from .guild import Guild
    from .message import Message
    from .partial_emoji import PartialEmoji
    from .role import Role
    from .state import ConnectionState
    from .types.activity import PresenceData
    from .types.gateway import GuildMemberUpdateEvent
    from .types.member import (
        BaseMember as BaseMemberPayload,
        Member as MemberPayload,
        MemberWithUser as MemberWithUserPayload,
        UserWithMember as UserWithMemberPayload,
    )
    from .types.user import AvatarDecorationData as AvatarDecorationDataPayload, User as UserPayload
    from .types.voice import (
        GuildVoiceState as GuildVoiceStatePayload,
        VoiceState as VoiceStatePayload,
    )

    VocalGuildChannel = Union[VoiceChannel, StageChannel]


class VoiceState:
    """Represents a Discord user's voice state.

    Attributes
    ----------
    deaf: :class:`bool`
        Whether the user is currently deafened by the guild.
    mute: :class:`bool`
        Whether the user is currently muted by the guild.
    self_mute: :class:`bool`
        Whether the user is currently muted by their own accord.
    self_deaf: :class:`bool`
        Whether the user is currently deafened by their own accord.
    self_stream: :class:`bool`
        Whether the user is currently streaming via 'Go Live' feature.

        .. versionadded:: 1.3

    self_video: :class:`bool`
        Whether the user is currently broadcasting video.
    suppress: :class:`bool`
        Whether the user is suppressed from speaking.

        Only applies to stage channels.

        .. versionadded:: 1.7

    requested_to_speak_at: Optional[:class:`datetime.datetime`]
        An aware datetime object that specifies the date and time in UTC that the member
        requested to speak. It will be ``None`` if they are not requesting to speak
        anymore or have been accepted to speak.

        Only applies to stage channels.

        .. versionadded:: 1.7

    afk: :class:`bool`
        Whether the user is currently in the AFK channel in the guild.
    channel: Optional[Union[:class:`VoiceChannel`, :class:`StageChannel`]]
        The voice channel that the user is currently connected to. ``None`` if the user
        is not currently in a voice channel.
    """

    __slots__ = (
        "session_id",
        "deaf",
        "mute",
        "self_mute",
        "self_stream",
        "self_video",
        "self_deaf",
        "afk",
        "channel",
        "requested_to_speak_at",
        "suppress",
    )

    def __init__(
        self,
        *,
        data: Union[VoiceStatePayload, GuildVoiceStatePayload],
        channel: Optional[VocalGuildChannel] = None,
    ) -> None:
        self.session_id: str = data["session_id"]
        self._update(data, channel)

    def _update(
        self,
        data: Union[VoiceStatePayload, GuildVoiceStatePayload],
        channel: Optional[VocalGuildChannel],
    ) -> None:
        self.self_mute: bool = data.get("self_mute", False)
        self.self_deaf: bool = data.get("self_deaf", False)
        self.self_stream: bool = data.get("self_stream", False)
        self.self_video: bool = data.get("self_video", False)
        self.afk: bool = data.get("suppress", False)
        self.mute: bool = data.get("mute", False)
        self.deaf: bool = data.get("deaf", False)
        self.suppress: bool = data.get("suppress", False)
        self.requested_to_speak_at: Optional[datetime.datetime] = utils.parse_time(
            data.get("request_to_speak_timestamp")
        )
        self.channel: Optional[VocalGuildChannel] = channel

    def __repr__(self) -> str:
        attrs = (
            ("self_mute", self.self_mute),
            ("self_deaf", self.self_deaf),
            ("self_stream", self.self_stream),
            ("suppress", self.suppress),
            ("requested_to_speak_at", self.requested_to_speak_at),
            ("channel", self.channel),
        )
        inner = " ".join(f"{k!s}={v!r}" for k, v in attrs)
        return f"<{self.__class__.__name__} {inner}>"


def flatten_user(cls):
    for attr, value in itertools.chain(BaseUser.__dict__.items(), User.__dict__.items()):
        # ignore private/special methods
        if attr.startswith("_"):
            continue

        # don't override what we already have
        if attr in cls.__dict__:
            continue

        # if it's a slotted attribute or a property, redirect it
        # slotted members are implemented as member_descriptors in Type.__dict__
        if not hasattr(value, "__annotations__"):
            getter = attrgetter(f"_user.{attr}")
            setattr(cls, attr, property(getter, doc=f"Equivalent to :attr:`User.{attr}`"))
        else:
            # Technically, this can also use attrgetter
            # However I'm not sure how I feel about "functions" returning properties
            # It probably breaks something in Sphinx.
            # probably a member function by now
            def generate_function(x):
                # We want sphinx to properly show coroutine functions as coroutines
                if asyncio.iscoroutinefunction(value):  # noqa: B023

                    async def general(self, *args, **kwargs):  # type: ignore
                        return await getattr(self._user, x)(*args, **kwargs)

                else:

                    def general(self, *args, **kwargs):
                        return getattr(self._user, x)(*args, **kwargs)

                general.__name__ = x
                return general

            func = generate_function(attr)
            func = utils.copy_doc(value)(func)
            setattr(cls, attr, func)

    return cls


@flatten_user
class Member(disnake.abc.Messageable, _UserTag):
    """Represents a Discord member to a :class:`Guild`.

    This implements a lot of the functionality of :class:`User`.

    .. collapse:: operations

        .. describe:: x == y

            Checks if two members are equal.
            Note that this works with :class:`User` instances too.

        .. describe:: x != y

            Checks if two members are not equal.
            Note that this works with :class:`User` instances too.

        .. describe:: hash(x)

            Returns the member's hash.

        .. describe:: str(x)

            Returns the member's username (with discriminator, if not migrated to new system yet).

    Attributes
    ----------
    joined_at: Optional[:class:`datetime.datetime`]
        An aware datetime object that specifies the date and time in UTC that the member joined the guild.
        If the member left and rejoined the guild, this will be the latest date. In certain cases, this can be ``None``.
    activities: Tuple[Union[:class:`BaseActivity`, :class:`Spotify`]]
        The activities that the user is currently doing.

        .. note::

            Due to a Discord API limitation, a user's Spotify activity may not appear
            if they are listening to a song with a title longer
            than 128 characters. See :issue-dpy:`1738` for more information.
    guild: :class:`Guild`
        The guild that the member belongs to.
    nick: Optional[:class:`str`]
        The guild specific nickname of the user.
        This takes precedence over :attr:`.global_name` and :attr:`.name` when shown.
    pending: :class:`bool`
        Whether the member is pending member verification.

        .. versionadded:: 1.6

    premium_since: Optional[:class:`datetime.datetime`]
        An aware datetime object that specifies the date and time in UTC when the member used their
        "Nitro boost" on the guild, if available. This could be ``None``.
    """

    __slots__ = (
        "_roles",
        "joined_at",
        "premium_since",
        "activities",
        "guild",
        "pending",
        "nick",
        "_client_status",
        "_user",
        "_state",
        "_avatar",
        "_banner",
        "_communication_disabled_until",
        "_flags",
        "_avatar_decoration_data",
    )

    if TYPE_CHECKING:
        name: str
        id: int
        discriminator: str
        global_name: Optional[str]
        bot: bool
        system: bool
        created_at: datetime.datetime
        default_avatar: Asset
        avatar: Optional[Asset]
        dm_channel: Optional[DMChannel]
        create_dm = User.create_dm
        mutual_guilds: List[Guild]
        public_flags: PublicUserFlags
        banner: Optional[Asset]
        accent_color: Optional[Colour]
        accent_colour: Optional[Colour]

    @overload
    def __init__(
        self,
        *,
        data: Union[MemberWithUserPayload, GuildMemberUpdateEvent],
        guild: Guild,
        state: ConnectionState,
    ) -> None:
        ...

    @overload
    def __init__(
        self,
        *,
        data: BaseMemberPayload,
        guild: Guild,
        state: ConnectionState,
        user_data: UserPayload,
    ) -> None:
        ...

    def __init__(
        self,
        *,
        data: Union[BaseMemberPayload, MemberWithUserPayload, GuildMemberUpdateEvent],
        guild: Guild,
        state: ConnectionState,
        user_data: Optional[UserPayload] = None,
    ) -> None:
        self._state: ConnectionState = state
        if user_data is None:
            user_data = cast("MemberWithUserPayload", data)["user"]
        self._user: User = state.store_user(user_data)
        self.guild: Guild = guild

        self.joined_at: Optional[datetime.datetime] = utils.parse_time(data.get("joined_at"))
        self.premium_since: Optional[datetime.datetime] = utils.parse_time(
            data.get("premium_since")
        )
        self._roles: utils.SnowflakeList = utils.SnowflakeList(map(int, data["roles"]))
        self._client_status: Dict[Optional[str], str] = {None: "offline"}
        self.activities: Tuple[ActivityTypes, ...] = ()
        self.nick: Optional[str] = data.get("nick")
        self.pending: bool = data.get("pending", False)
        self._avatar: Optional[str] = data.get("avatar")
        self._banner: Optional[str] = data.get("banner")
        timeout_datetime = utils.parse_time(data.get("communication_disabled_until"))
        self._communication_disabled_until: Optional[datetime.datetime] = timeout_datetime
        self._flags: int = data.get("flags", 0)
        self._avatar_decoration_data: Optional[AvatarDecorationDataPayload] = data.get(
            "avatar_decoration_data"
        )

    def __str__(self) -> str:
        return str(self._user)

    def __repr__(self) -> str:
        return (
            f"<Member id={self._user.id} name={self._user.name!r} global_name={self._user.global_name!r} discriminator={self._user.discriminator!r}"
            f" bot={self._user.bot} nick={self.nick!r} guild={self.guild!r}>"
        )

    def __eq__(self, other: Any) -> bool:
        return isinstance(other, _UserTag) and other.id == self.id

    def __ne__(self, other: Any) -> bool:
        return not self.__eq__(other)

    def __hash__(self) -> int:
        return hash(self._user)

    @classmethod
    def _from_message(cls, *, message: Message, data: MemberPayload) -> Self:
        user_data = message.author._to_minimal_user_json()  # type: ignore
        return cls(
            data=data,
            user_data=user_data,
            guild=message.guild,  # type: ignore
            state=message._state,
        )

    def _update_from_message(self, data: MemberPayload) -> None:
        self.joined_at = utils.parse_time(data.get("joined_at"))
        self.premium_since = utils.parse_time(data.get("premium_since"))
        self._roles = utils.SnowflakeList(map(int, data["roles"]))
        self.nick = data.get("nick", None)
        self.pending = data.get("pending", False)
        self._flags = data.get("flags", 0)

    @classmethod
    def _try_upgrade(
        cls,
        *,
        data: Union[UserPayload, UserWithMemberPayload],
        guild: Guild,
        state: ConnectionState,
    ) -> Union[User, Self]:
        # A User object with a 'member' key
        try:
            member_data = data.pop("member")  # type: ignore
        except KeyError:
            return state.create_user(data)
        else:
            return cls(data=member_data, user_data=data, guild=guild, state=state)

    @classmethod
    def _copy(cls, member: Member) -> Self:
        self = cls.__new__(cls)  # to bypass __init__

        self._roles = utils.SnowflakeList(member._roles, is_sorted=True)
        self.joined_at = member.joined_at
        self.premium_since = member.premium_since
        self._client_status = member._client_status.copy()
        self.guild = member.guild
        self.nick = member.nick
        self.pending = member.pending
        self.activities = member.activities
        self._state = member._state
        self._avatar = member._avatar
        self._banner = member._banner
        self._communication_disabled_until = member.current_timeout
        self._flags = member._flags

        # Reference will not be copied unless necessary by PRESENCE_UPDATE
        # See below
        self._user = member._user
        return self

    async def _get_channel(self):
        ch = await self.create_dm()
        return ch

    def _update(self, data: GuildMemberUpdateEvent) -> None:
        # the nickname change is optional,
        # if it isn't in the payload then it didn't change
        try:
            self.nick = data["nick"]
        except KeyError:
            pass

        try:
            self.pending = data["pending"]
        except KeyError:
            pass

        self.premium_since = utils.parse_time(data.get("premium_since"))
        self._roles = utils.SnowflakeList(map(int, data["roles"]))
        self._avatar = data.get("avatar")
        self._banner = data.get("banner")
        timeout_datetime = utils.parse_time(data.get("communication_disabled_until"))
        self._communication_disabled_until = timeout_datetime
        self._flags = data.get("flags", 0)
        self._avatar_decoration_data = data.get("avatar_decoration_data")

    def _presence_update(
        self, data: PresenceData, user: UserPayload
    ) -> Optional[Tuple[User, User]]:
        self.activities = tuple(create_activity(a, state=self._state) for a in data["activities"])
        self._client_status = {
            sys.intern(key): sys.intern(value) for key, value in data.get("client_status", {}).items()  # type: ignore
        }
        self._client_status[None] = sys.intern(data["status"])

        if len(user) > 1:
            return self._update_inner_user(user)
        return None

    def _update_inner_user(self, user: UserPayload) -> Optional[Tuple[User, User]]:
        u = self._user
        original = (
            u.name,
            u._avatar,
            u.discriminator,
            u.global_name,
            u._public_flags,
            u._avatar_decoration_data,
        )
        # These keys seem to always be available
        modified = (
            user["username"],
            user["avatar"],
            user["discriminator"],
            user.get("global_name"),
            user.get("public_flags", 0),
            user.get("avatar_decoration_data", None),
        )
        if original != modified:
            to_return = User._copy(self._user)
            (
                u.name,
                u._avatar,
                u.discriminator,
                u.global_name,
                u._public_flags,
                u._avatar_decoration_data,
            ) = modified
            # Signal to dispatch on_user_update
            return to_return, u

    @property
    def status(self) -> Status:
        """:class:`Status`: The member's overall status. If the value is unknown, then it will be a :class:`str` instead."""
        return try_enum(Status, self._client_status[None])

    @property
    def raw_status(self) -> str:
        """:class:`str`: The member's overall status as a string value.

        .. versionadded:: 1.5
        """
        return self._client_status[None]

    @status.setter
    def status(self, value: Status) -> None:
        # internal use only
        self._client_status[None] = str(value)

    @property
    def tag(self) -> str:
        """:class:`str`: An alias of :attr:`.discriminator`."""
        return self.discriminator

    @property
    def mobile_status(self) -> Status:
        """:class:`Status`: The member's status on a mobile device, if applicable."""
        return try_enum(Status, self._client_status.get("mobile", "offline"))

    @property
    def desktop_status(self) -> Status:
        """:class:`Status`: The member's status on the desktop client, if applicable."""
        return try_enum(Status, self._client_status.get("desktop", "offline"))

    @property
    def web_status(self) -> Status:
        """:class:`Status`: The member's status on the web client, if applicable."""
        return try_enum(Status, self._client_status.get("web", "offline"))

    def is_on_mobile(self) -> bool:
        """Whether the member is active on a mobile device.

        :return type: :class:`bool`
        """
        return "mobile" in self._client_status

    @property
    def colour(self) -> Colour:
        """:class:`Colour`: A property that returns a colour denoting the rendered colour
        for the member. If the default colour is the one rendered then an instance
        of :meth:`Colour.default` is returned.

        There is an alias for this named :attr:`color`.
        """
        roles = self.roles[1:]  # remove @everyone

        # highest order of the colour is the one that gets rendered.
        # if the highest is the default colour then the next one with a colour
        # is chosen instead
        for role in reversed(roles):
            if role.colour.value:
                return role.colour
        return Colour.default()

    @property
    def color(self) -> Colour:
        """:class:`Colour`: A property that returns a color denoting the rendered color for
        the member. If the default color is the one rendered then an instance of :meth:`Colour.default`
        is returned.

        There is an alias for this named :attr:`colour`.
        """
        return self.colour

    @property
    def roles(self) -> List[Role]:
        """List[:class:`Role`]: A :class:`list` of :class:`Role` that the member belongs to. Note
        that the first element of this list is always the default '@everyone'
        role.

        These roles are sorted by their position in the role hierarchy.
        """
        result = []
        g = self.guild
        for role_id in self._roles:
            role = g.get_role(role_id)
            if role:
                result.append(role)
        result.append(g.default_role)
        result.sort()
        return result

    @property
    def mention(self) -> str:
        """:class:`str`: Returns a string that allows you to mention the member."""
        return f"<@{self._user.id}>"

    @property
    def display_name(self) -> str:
        """:class:`str`: Returns the user's display name.

        If they have a guild-specific :attr:`nickname <.nick>`, then
        that is returned. If not, this is their :attr:`global name <.global_name>`
        if set, or their :attr:`username <.name>` otherwise.

        .. versionchanged:: 2.9
            Added :attr:`.global_name`.
        """
        return self.nick or self.global_name or self.name

    @property
    def display_avatar(self) -> Asset:
        """:class:`Asset`: Returns the member's display avatar.

        For regular members this is just their avatar, but
        if they have a guild specific avatar then that
        is returned instead.

        .. versionadded:: 2.0
        """
        return self.guild_avatar or self._user.avatar or self._user.default_avatar

    @property
    def guild_avatar(self) -> Optional[Asset]:
        """Optional[:class:`Asset`]: Returns an :class:`Asset` for the guild avatar
        the member has. If unavailable, ``None`` is returned.

        .. versionadded:: 2.0
        """
        if self._avatar is None:
            return None
        return Asset._from_guild_avatar(self._state, self.guild.id, self.id, self._avatar)

    # TODO
    # implement a `display_banner` property
    # for more info on why this wasn't implemented read this discussion
    # https://github.com/DisnakeDev/disnake/pull/1204#discussion_r1685773429
    @property
    def guild_banner(self) -> Optional[Asset]:
        """Optional[:class:`Asset`]: Returns an :class:`Asset` for the guild banner
        the member has. If unavailable, ``None`` is returned.

        .. versionadded:: 2.10
        """
        if self._banner is None:
            return None
        return Asset._from_guild_banner(self._state, self.guild.id, self.id, self._banner)

    @property
    def activity(self) -> Optional[ActivityTypes]:
        """Optional[Union[:class:`BaseActivity`, :class:`Spotify`]]: Returns the primary
        activity the user is currently doing. Could be ``None`` if no activity is being done.

        .. note::

            Due to a Discord API limitation, this may be ``None`` if
            the user is listening to a song on Spotify with a title longer
            than 128 characters. See :issue-dpy:`1738` for more information.

        .. note::

            A user may have multiple activities, these can be accessed under :attr:`activities`.
        """
        if self.activities:
            return self.activities[0]

    def mentioned_in(self, message: Message) -> bool:
        """Whether the member is mentioned in the specified message.

        Parameters
        ----------
        message: :class:`Message`
            The message to check.

        Returns
        -------
        :class:`bool`
            Indicates if the member is mentioned in the message.
        """
        if message.guild is None or message.guild.id != self.guild.id:
            return False

        if self._user.mentioned_in(message):
            return True

        return any(self._roles.has(role.id) for role in message.role_mentions)

    @property
    def top_role(self) -> Role:
        """:class:`Role`: Returns the member's highest role.

        This is useful for figuring where a member stands in the role
        hierarchy chain.
        """
        guild = self.guild
        if len(self._roles) == 0:
            return guild.default_role

        return max(guild.get_role(rid) or guild.default_role for rid in self._roles)

    @property
    def role_icon(self) -> Optional[Union[Asset, PartialEmoji]]:
        """Optional[Union[:class:`Asset`, :class:`PartialEmoji`]]: Returns the member's displayed role icon, if any.

        .. versionadded:: 2.5
        """
        roles = self.roles[1:]  # remove @everyone

        for role in reversed(roles):
            if icon := (role.icon or role.emoji):
                return icon
        return None

    @property
    def guild_permissions(self) -> Permissions:
        """:class:`Permissions`: Returns the member's guild permissions.

        This only takes into consideration the guild permissions
        and not most of the implied permissions or any of the
        channel permission overwrites. For 100% accurate permission
        calculation, please use :meth:`abc.GuildChannel.permissions_for`.

        This does take into consideration guild ownership and the
        administrator implication.
        """
        if self.guild.owner_id == self.id:
            return Permissions.all()

        base = Permissions.none()
        for r in self.roles:
            base.value |= r.permissions.value

        if base.administrator:
            return Permissions.all()

        return base

    @property
    def voice(self) -> Optional[VoiceState]:
        """Optional[:class:`VoiceState`]: Returns the member's current voice state."""
        return self.guild._voice_state_for(self._user.id)

    @property
    def current_timeout(self) -> Optional[datetime.datetime]:
        """Optional[:class:`datetime.datetime`]: Returns the datetime when the timeout expires.

        If the member is not timed out or the timeout has already expired, returns ``None``.

        .. versionadded:: 2.3
        """
        if self._communication_disabled_until is None:
            return None

        if self._communication_disabled_until < utils.utcnow():
            self._communication_disabled_until = None
            return None

        return self._communication_disabled_until

    @property
    def flags(self) -> MemberFlags:
        """:class:`MemberFlags`: Returns the member's flags.

        .. versionadded:: 2.8
        """
        return MemberFlags._from_value(self._flags)

    @property
    def display_avatar_decoration(self) -> Optional[Asset]:
        """Optional[:class:`Asset`]: Returns the member's display avatar decoration.

        For regular members this is just their avatar decoration, but
        if they have a guild specific avatar decoration then that
        is returned instead.

        .. versionadded:: 2.10

        .. note::

            Since Discord always sends an animated PNG for animated avatar decorations,
            the following methods will not work as expected:

            - :meth:`Asset.replace`
            - :meth:`Asset.with_size`
            - :meth:`Asset.with_format`
            - :meth:`Asset.with_static_format`
        """
        return self.guild_avatar_decoration or self._user.avatar_decoration

    @property
    def guild_avatar_decoration(self) -> Optional[Asset]:
        """Optional[:class:`Asset`]: Returns an :class:`Asset` for the guild avatar decoration
        the member has. If unavailable, ``None`` is returned.

        .. versionadded:: 2.10

        .. note::

            Since Discord always sends an animated PNG for animated avatar decorations,
            the following methods will not work as expected:

            - :meth:`Asset.replace`
            - :meth:`Asset.with_size`
            - :meth:`Asset.with_format`
            - :meth:`Asset.with_static_format`
        """
        if self._avatar_decoration_data is None:
            return None
        return Asset._from_avatar_decoration(self._state, self._avatar_decoration_data["asset"])

    @overload
    async def ban(
        self,
        *,
        clean_history_duration: Union[int, datetime.timedelta] = 86400,
        reason: Optional[str] = None,
    ) -> None:
        ...

    @overload
    async def ban(
        self,
        *,
        delete_message_days: Literal[0, 1, 2, 3, 4, 5, 6, 7] = 1,
        reason: Optional[str] = None,
    ) -> None:
        ...

    async def ban(
        self,
        *,
        clean_history_duration: Union[int, datetime.timedelta] = MISSING,
        delete_message_days: Literal[0, 1, 2, 3, 4, 5, 6, 7] = MISSING,
        reason: Optional[str] = None,
    ) -> None:
        """|coro|

        Bans this member. Equivalent to :meth:`Guild.ban`.
        """
        await self.guild.ban(  # type: ignore  # no matching overload
            self,
            reason=reason,
            clean_history_duration=clean_history_duration,
            delete_message_days=delete_message_days,
        )

    async def unban(self, *, reason: Optional[str] = None) -> None:
        """|coro|

        Unbans this member. Equivalent to :meth:`Guild.unban`.
        """
        await self.guild.unban(self, reason=reason)

    async def kick(self, *, reason: Optional[str] = None) -> None:
        """|coro|

        Kicks this member. Equivalent to :meth:`Guild.kick`.
        """
        await self.guild.kick(self, reason=reason)

    async def edit(
        self,
        *,
        nick: Optional[str] = MISSING,
        mute: bool = MISSING,
        deafen: bool = MISSING,
        suppress: bool = MISSING,
        roles: Sequence[disnake.abc.Snowflake] = MISSING,
        voice_channel: Optional[VocalGuildChannel] = MISSING,
        timeout: Optional[Union[float, datetime.timedelta, datetime.datetime]] = MISSING,
        flags: MemberFlags = MISSING,
        bypasses_verification: bool = MISSING,
        reason: Optional[str] = None,
    ) -> Optional[Member]:
        """|coro|

        Edits the member's data.

        Depending on the parameter passed, this requires different permissions listed below:

        +------------------------------+--------------------------------------+
        |   Parameter                  |              Permission              |
        +==============================+======================================+
        | nick                         | :attr:`Permissions.manage_nicknames` |
        +------------------------------+--------------------------------------+
        | mute                         | :attr:`Permissions.mute_members`     |
        +------------------------------+--------------------------------------+
        | deafen                       | :attr:`Permissions.deafen_members`   |
        +------------------------------+--------------------------------------+
        | roles                        | :attr:`Permissions.manage_roles`     |
        +------------------------------+--------------------------------------+
        | voice_channel                | :attr:`Permissions.move_members`     |
        +------------------------------+--------------------------------------+
        | timeout                      | :attr:`Permissions.moderate_members` |
        +------------------------------+--------------------------------------+
        | flags                        | :attr:`Permissions.manage_guild` or  |
        |                              | :attr:`Permissions.manage_roles` or  |
        |                              | (:attr:`Permissions.moderate_members`|
        |                              | + :attr:`Permissions.kick_members`   |
        |                              | + :attr:`Permissions.ban_members`)   |
        +------------------------------+--------------------------------------+
        | bypasses_verification        | (same as ``flags``)                  |
        +------------------------------+--------------------------------------+

        All parameters are optional.

        .. versionchanged:: 1.1
            Can now pass ``None`` to ``voice_channel`` to kick a member from voice.

        .. versionchanged:: 2.0
            The newly member is now optionally returned, if applicable.

        Parameters
        ----------
        nick: Optional[:class:`str`]
            The member's new nickname. Use ``None`` to remove the nickname.
        mute: :class:`bool`
            Whether the member should be guild muted or un-muted.
        deafen: :class:`bool`
            Whether the member should be guild deafened or un-deafened.
        suppress: :class:`bool`
            Whether the member should be suppressed in stage channels.

            .. versionadded:: 1.7

        roles: Sequence[:class:`Role`]
            The member's new list of roles. This *replaces* the roles.
        voice_channel: Optional[:class:`VoiceChannel`]
            The voice channel to move the member to.
            Pass ``None`` to kick them from voice.
        timeout: Optional[Union[:class:`float`, :class:`datetime.timedelta`, :class:`datetime.datetime`]]
            The duration (seconds or timedelta) or the expiry (datetime) of the timeout;
            until then, the member will not be able to interact with the guild.
            Set to ``None`` to remove the timeout. Supports up to 28 days in the future.

            .. versionadded:: 2.3

        flags: :class:`MemberFlags`
            The member's new flags. To know what flags are editable,
            see :ddocs:`the documentation <resources/guild#guild-member-object-guild-member-flags>`.

            If parameter ``bypasses_verification`` is provided, that will override the setting of :attr:`MemberFlags.bypasses_verification`.

            .. versionadded:: 2.8

        bypasses_verification: :class:`bool`
            Whether the member bypasses guild verification requirements.

            .. versionadded:: 2.8

        reason: Optional[:class:`str`]
            The reason for editing this member. Shows up on the audit log.

        Raises
        ------
        Forbidden
            You do not have the proper permissions to the action requested.
        HTTPException
            The operation failed.

        Returns
        -------
        Optional[:class:`.Member`]
            The newly updated member, if applicable. This is only returned
            when certain fields are updated.
        """
        http = self._state.http
        guild_id = self.guild.id
        me = self._state.self_id == self.id
        payload: Dict[str, Any] = {}

        if nick is not MISSING:
            nick = nick or ""
            if me:
                await http.edit_my_member(guild_id, nick=nick, reason=reason)
            else:
                payload["nick"] = nick

        if deafen is not MISSING:
            payload["deaf"] = deafen

        if mute is not MISSING:
            payload["mute"] = mute

        if suppress is not MISSING:
            if self.voice is None or self.voice.channel is None:
                raise Exception("Cannot suppress a member which isn't in a vc")  # noqa: TRY002

            voice_state_payload: Dict[str, Any] = {
                "channel_id": self.voice.channel.id,
                "suppress": suppress,
            }

            if suppress or self.bot:
                voice_state_payload["request_to_speak_timestamp"] = None

            if me:
                await http.edit_my_voice_state(guild_id, voice_state_payload)
            else:
                if not suppress:
                    voice_state_payload["request_to_speak_timestamp"] = utils.utcnow().isoformat()
                await http.edit_voice_state(guild_id, self.id, voice_state_payload)

        if voice_channel is not MISSING:
            payload["channel_id"] = voice_channel and voice_channel.id

        if roles is not MISSING:
            payload["roles"] = tuple(r.id for r in roles)

        if timeout is not MISSING:
            if timeout is not None:
                if isinstance(timeout, datetime.datetime):
                    dt = timeout.astimezone(tz=datetime.timezone.utc)
                elif isinstance(timeout, datetime.timedelta):
                    dt = utils.utcnow() + timeout
                else:
                    dt = utils.utcnow() + datetime.timedelta(seconds=timeout)
                payload["communication_disabled_until"] = dt.isoformat()
            else:
                payload["communication_disabled_until"] = None

        if bypasses_verification is not MISSING:
            # create base flags if flags are provided, otherwise use the internal flags.
            flags = MemberFlags._from_value(self._flags if flags is MISSING else flags.value)
            flags.bypasses_verification = bypasses_verification

        if flags is not MISSING:
            payload["flags"] = flags.value

        if payload:
            data = await http.edit_member(guild_id, self.id, reason=reason, **payload)
            return Member(data=data, guild=self.guild, state=self._state)

    async def request_to_speak(self) -> None:
        """|coro|

        Requests to speak in the connected channel.

        Only applies to stage channels.

        .. note::

            Requesting members that are not the client is equivalent
            to :attr:`.edit` providing ``suppress`` as ``False``.

        .. versionadded:: 1.7

        Raises
        ------
        Forbidden
            You do not have the proper permissions to the action requested.
        HTTPException
            The operation failed.
        """
        if self.voice is None or self.voice.channel is None:
            raise Exception("Cannot request to speak when not in a vc")  # noqa: TRY002

        payload = {
            "channel_id": self.voice.channel.id,
            "request_to_speak_timestamp": utils.utcnow().isoformat(),
        }

        if self._state.self_id != self.id:
            payload["suppress"] = False
            await self._state.http.edit_voice_state(self.guild.id, self.id, payload)
        else:
            await self._state.http.edit_my_voice_state(self.guild.id, payload)

    async def move_to(self, channel: VocalGuildChannel, *, reason: Optional[str] = None) -> None:
        """|coro|

        Moves a member to a new voice channel (they must be connected first).

        You must have :attr:`~Permissions.move_members` permission to
        use this.

        This raises the same exceptions as :meth:`edit`.

        .. versionchanged:: 1.1
            Can now pass ``None`` to kick a member from voice.

        Parameters
        ----------
        channel: Optional[:class:`VoiceChannel`]
            The new voice channel to move the member to.
            Pass ``None`` to kick them from voice.
        reason: Optional[:class:`str`]
            The reason for doing this action. Shows up on the audit log.
        """
        await self.edit(voice_channel=channel, reason=reason)

    async def add_roles(
        self, *roles: Snowflake, reason: Optional[str] = None, atomic: bool = True
    ) -> None:
        """|coro|

        Gives the member a number of :class:`Role`\\s.

        You must have :attr:`~Permissions.manage_roles` permission to
        use this, and the added :class:`Role`\\s must appear lower in the list
        of roles than the highest role of the member.

        Parameters
        ----------
        *roles: :class:`abc.Snowflake`
            An argument list of :class:`abc.Snowflake` representing a :class:`Role`
            to give to the member.
        reason: Optional[:class:`str`]
            The reason for adding these roles. Shows up on the audit log.
        atomic: :class:`bool`
            Whether to atomically add roles. This will ensure that multiple
            operations will always be applied regardless of the current
            state of the cache.

        Raises
        ------
        Forbidden
            You do not have permissions to add these roles.
        HTTPException
            Adding roles failed.
        """
        if not atomic:
            new_roles = utils._unique(Object(id=r.id) for s in (self.roles[1:], roles) for r in s)
            await self.edit(roles=new_roles, reason=reason)
        else:
            req = self._state.http.add_role
            guild_id = self.guild.id
            user_id = self.id
            for role in roles:
                await req(guild_id, user_id, role.id, reason=reason)

    async def remove_roles(
        self, *roles: Snowflake, reason: Optional[str] = None, atomic: bool = True
    ) -> None:
        """|coro|

        Removes :class:`Role`\\s from this member.

        You must have :attr:`~Permissions.manage_roles` permission to
        use this, and the removed :class:`Role`\\s must appear lower in the list
        of roles than the highest role of the member.

        Parameters
        ----------
        *roles: :class:`abc.Snowflake`
            An argument list of :class:`abc.Snowflake` representing a :class:`Role`
            to remove from the member.
        reason: Optional[:class:`str`]
            The reason for removing these roles. Shows up on the audit log.
        atomic: :class:`bool`
            Whether to atomically remove roles. This will ensure that multiple
            operations will always be applied regardless of the current
            state of the cache.

        Raises
        ------
        Forbidden
            You do not have permissions to remove these roles.
        HTTPException
            Removing the roles failed.
        """
        if not atomic:
            new_roles = [Object(id=r.id) for r in self.roles[1:]]  # remove @everyone
            for role in roles:
                try:
                    new_roles.remove(Object(id=role.id))
                except ValueError:
                    pass

            await self.edit(roles=new_roles, reason=reason)
        else:
            req = self._state.http.remove_role
            guild_id = self.guild.id
            user_id = self.id
            for role in roles:
                await req(guild_id, user_id, role.id, reason=reason)

    def get_role(self, role_id: int, /) -> Optional[Role]:
        """Returns a role with the given ID from roles which the member has.

        .. versionadded:: 2.0

        Parameters
        ----------
        role_id: :class:`int`
            The role ID to search for.

        Returns
        -------
        Optional[:class:`Role`]
            The role or ``None`` if not found in the member's roles.
        """
        return self.guild.get_role(role_id) if self._roles.has(role_id) else None

    @overload
    async def timeout(
        self,
        *,
        duration: Optional[Union[float, datetime.timedelta]],
        reason: Optional[str] = None,
    ) -> Member:
        ...

    @overload
    async def timeout(
        self,
        *,
        until: Optional[datetime.datetime],
        reason: Optional[str] = None,
    ) -> Member:
        ...

    async def timeout(
        self,
        *,
        duration: Optional[Union[float, datetime.timedelta]] = MISSING,
        until: Optional[datetime.datetime] = MISSING,
        reason: Optional[str] = None,
    ) -> Member:
        """|coro|

        Times out the member from the guild; until then, the member will not be able to interact with the guild.

        Exactly one of ``duration`` or ``until`` must be provided. To remove a timeout, set one of the parameters to ``None``.

        You must have the :attr:`Permissions.moderate_members` permission to do this.

        .. versionadded:: 2.3

        Parameters
        ----------
        duration: Optional[Union[:class:`float`, :class:`datetime.timedelta`]]
            The duration (seconds or timedelta) of the member's timeout. Set to ``None`` to remove the timeout.
            Supports up to 28 days in the future.
            May not be used in combination with the ``until`` parameter.
        until: Optional[:class:`datetime.datetime`]
            The expiry date/time of the member's timeout. Set to ``None`` to remove the timeout.
            Supports up to 28 days in the future.
            May not be used in combination with the ``duration`` parameter.
        reason: Optional[:class:`str`]
            The reason for this timeout. Appears on the audit log.

        Raises
        ------
        Forbidden
            You do not have permissions to timeout this member.
        HTTPException
            Timing out the member failed.

        Returns
        -------
        :class:`Member`
            The newly updated member.
        """
        if duration is not MISSING:
            return await self.guild.timeout(self, duration=duration, reason=reason)
        else:
            return await self.guild.timeout(self, until=until, reason=reason)



# SPDX-License-Identifier: MIT

from __future__ import annotations

import datetime
from typing import TYPE_CHECKING, List, Literal, Optional, Set, Union, cast

from .enums import ChannelType, try_enum
from .utils import _get_as_snowflake, get_slots

if TYPE_CHECKING:
    from .channel import VoiceChannelEffect
    from .member import Member
    from .message import Message
    from .partial_emoji import PartialEmoji
    from .threads import Thread, ThreadMember, ThreadType
    from .types.gateway import (
        GuildScheduledEventUserAddEvent,
        GuildScheduledEventUserRemoveEvent,
        IntegrationDeleteEvent,
        MessageDeleteBulkEvent,
        MessageDeleteEvent,
        MessageReactionAddEvent,
        MessageReactionRemoveAllEvent,
        MessageReactionRemoveEmojiEvent,
        MessageReactionRemoveEvent,
        MessageUpdateEvent,
        PollVoteAddEvent,
        PollVoteRemoveEvent,
        PresenceUpdateEvent,
        ThreadDeleteEvent,
        TypingStartEvent,
        VoiceChannelEffectSendEvent,
    )
    from .user import User


__all__ = (
    "RawMessageDeleteEvent",
    "RawBulkMessageDeleteEvent",
    "RawMessageUpdateEvent",
    "RawReactionActionEvent",
    "RawReactionClearEvent",
    "RawReactionClearEmojiEvent",
    "RawIntegrationDeleteEvent",
    "RawGuildScheduledEventUserActionEvent",
    "RawThreadDeleteEvent",
    "RawThreadMemberRemoveEvent",
    "RawTypingEvent",
    "RawGuildMemberRemoveEvent",
    "RawPresenceUpdateEvent",
    "RawPollVoteActionEvent",
    "RawVoiceChannelEffectEvent",
)


class _RawReprMixin:
    __slots__ = ()

    def __repr__(self) -> str:
        value = " ".join(f"{attr}={getattr(self, attr)!r}" for attr in get_slots(type(self)))
        return f"<{type(self).__name__} {value}>"


class RawMessageDeleteEvent(_RawReprMixin):
    """Represents the event payload for an :func:`on_raw_message_delete` event.

    Attributes
    ----------
    channel_id: :class:`int`
        The channel ID where the deletion took place.
    guild_id: Optional[:class:`int`]
        The guild ID where the deletion took place, if applicable.
    message_id: :class:`int`
        The message ID that got deleted.
    cached_message: Optional[:class:`Message`]
        The cached message, if found in the internal message cache.
    """

    __slots__ = ("message_id", "channel_id", "guild_id", "cached_message")

    def __init__(self, data: MessageDeleteEvent) -> None:
        self.message_id: int = int(data["id"])
        self.channel_id: int = int(data["channel_id"])
        self.cached_message: Optional[Message] = None
        try:
            self.guild_id: Optional[int] = int(data["guild_id"])
        except KeyError:
            self.guild_id: Optional[int] = None


class RawBulkMessageDeleteEvent(_RawReprMixin):
    """Represents the event payload for an :func:`on_raw_bulk_message_delete` event.

    Attributes
    ----------
    message_ids: Set[:class:`int`]
        A :class:`set` of the message IDs that were deleted.
    channel_id: :class:`int`
        The channel ID where the deletion took place.
    guild_id: Optional[:class:`int`]
        The guild ID where the deletion took place, if applicable.
    cached_messages: List[:class:`Message`]
        The cached messages, if found in the internal message cache.
    """

    __slots__ = ("message_ids", "channel_id", "guild_id", "cached_messages")

    def __init__(self, data: MessageDeleteBulkEvent) -> None:
        self.message_ids: Set[int] = {int(x) for x in data.get("ids", [])}
        self.channel_id: int = int(data["channel_id"])
        self.cached_messages: List[Message] = []
        try:
            self.guild_id: Optional[int] = int(data["guild_id"])
        except KeyError:
            self.guild_id: Optional[int] = None


class RawMessageUpdateEvent(_RawReprMixin):
    """Represents the event payload for an :func:`on_raw_message_edit` event.

    Attributes
    ----------
    message_id: :class:`int`
        The message ID that got updated.
    channel_id: :class:`int`
        The channel ID where the update took place.

        .. versionadded:: 1.3

    guild_id: Optional[:class:`int`]
        The guild ID where the update took place, if applicable.

        .. versionadded:: 1.7

    data: :class:`dict`
        The raw data given by the :ddocs:`gateway <topics/gateway-events#message-update>`.
    cached_message: Optional[:class:`Message`]
        The cached message, if found in the internal message cache. Represents the message before
        it is modified by the data in :attr:`RawMessageUpdateEvent.data`.
    """

    __slots__ = ("message_id", "channel_id", "guild_id", "data", "cached_message")

    def __init__(self, data: MessageUpdateEvent) -> None:
        self.message_id: int = int(data["id"])
        self.channel_id: int = int(data["channel_id"])
        self.data: MessageUpdateEvent = data
        self.cached_message: Optional[Message] = None
        try:
            self.guild_id: Optional[int] = int(data["guild_id"])
        except KeyError:
            self.guild_id: Optional[int] = None


PollEventType = Literal["POLL_VOTE_ADD", "POLL_VOTE_REMOVE"]


class RawPollVoteActionEvent(_RawReprMixin):
    """Represents the event payload for :func:`on_raw_poll_vote_add` and
    :func:`on_raw_poll_vote_remove` events.

    .. versionadded:: 2.10

    Attributes
    ----------
    message_id: :class:`int`
        The message ID that got or lost a vote.
    user_id: :class:`int`
        The user ID who added the vote or whose vote was removed.
    cached_member: Optional[:class:`Member`]
        The member who added the vote. Available only when the guilds and members are cached.
    channel_id: :class:`int`
        The channel ID where the vote addition or removal took place.
    guild_id: Optional[:class:`int`]
        The guild ID where the vote addition or removal took place, if applicable.
    answer_id: :class:`int`
        The ID of the answer that was voted or unvoted.
    event_type: :class:`str`
        The event type that triggered this action. Can be
        ``POLL_VOTE_ADD`` for vote addition or
        ``POLL_VOTE_REMOVE`` for vote removal.
    """

    __slots__ = (
        "message_id",
        "user_id",
        "cached_member",
        "channel_id",
        "guild_id",
        "event_type",
        "answer_id",
    )

    def __init__(
        self,
        data: Union[PollVoteAddEvent, PollVoteRemoveEvent],
        event_type: PollEventType,
    ) -> None:
        self.message_id: int = int(data["message_id"])
        self.user_id: int = int(data["user_id"])
        self.cached_member: Optional[Member] = None
        self.channel_id: int = int(data["channel_id"])
        self.event_type = event_type
        self.answer_id: int = int(data["answer_id"])
        try:
            self.guild_id: Optional[int] = int(data["guild_id"])
        except KeyError:
            self.guild_id: Optional[int] = None


ReactionEventType = Literal["REACTION_ADD", "REACTION_REMOVE"]


class RawReactionActionEvent(_RawReprMixin):
    """Represents the event payload for :func:`on_raw_reaction_add` and
    :func:`on_raw_reaction_remove` events.

    Attributes
    ----------
    message_id: :class:`int`
        The message ID that got or lost a reaction.
    user_id: :class:`int`
        The user ID who added the reaction or whose reaction was removed.
    channel_id: :class:`int`
        The channel ID where the reaction addition or removal took place.
    guild_id: Optional[:class:`int`]
        The guild ID where the reaction addition or removal took place, if applicable.
    emoji: :class:`PartialEmoji`
        The custom or unicode emoji being used.

        .. versionchanged:: 2.9
            This now also includes the correct :attr:`~PartialEmoji.animated` value when a reaction was removed.

    member: Optional[:class:`Member`]
        The member who added the reaction. Only available if :attr:`event_type` is ``REACTION_ADD`` and the reaction is inside a guild.

        .. versionadded:: 1.3

    event_type: :class:`str`
        The event type that triggered this action. Can be
        ``REACTION_ADD`` for reaction addition or
        ``REACTION_REMOVE`` for reaction removal.

        .. versionadded:: 1.3

    message_author_id: Optional[:class:`int`]
        The ID of the author who created the message that got a reaction.
        Only available if :attr:`event_type` is ``REACTION_ADD``.
        May also be ``None`` if the message was created by a webhook.

        .. versionadded:: 2.10
    """

    __slots__ = (
        "message_id",
        "user_id",
        "channel_id",
        "guild_id",
        "emoji",
        "event_type",
        "member",
        "message_author_id",
    )

    def __init__(
        self,
        data: Union[MessageReactionAddEvent, MessageReactionRemoveEvent],
        emoji: PartialEmoji,
        event_type: ReactionEventType,
    ) -> None:
        self.message_id: int = int(data["message_id"])
        self.channel_id: int = int(data["channel_id"])
        self.user_id: int = int(data["user_id"])
        self.emoji: PartialEmoji = emoji
        self.event_type: ReactionEventType = event_type
        self.member: Optional[Member] = None
        try:
            self.guild_id: Optional[int] = int(data["guild_id"])
        except KeyError:
            self.guild_id: Optional[int] = None
        self.message_author_id: Optional[int] = _get_as_snowflake(data, "message_author_id")


class RawReactionClearEvent(_RawReprMixin):
    """Represents the event payload for an :func:`on_raw_reaction_clear` event.

    Attributes
    ----------
    message_id: :class:`int`
        The message ID that got its reactions cleared.
    channel_id: :class:`int`
        The channel ID where the reaction clear took place.
    guild_id: Optional[:class:`int`]
        The guild ID where the reaction clear took place, if applicable.
    """

    __slots__ = ("message_id", "channel_id", "guild_id")

    def __init__(self, data: MessageReactionRemoveAllEvent) -> None:
        self.message_id: int = int(data["message_id"])
        self.channel_id: int = int(data["channel_id"])
        try:
            self.guild_id: Optional[int] = int(data["guild_id"])
        except KeyError:
            self.guild_id: Optional[int] = None


class RawReactionClearEmojiEvent(_RawReprMixin):
    """Represents the event payload for an :func:`on_raw_reaction_clear_emoji` event.

    .. versionadded:: 1.3

    Attributes
    ----------
    message_id: :class:`int`
        The message ID that got its reactions cleared.
    channel_id: :class:`int`
        The channel ID where the reaction clear took place.
    guild_id: Optional[:class:`int`]
        The guild ID where the reaction clear took place, if applicable.
    emoji: :class:`PartialEmoji`
        The custom or unicode emoji being removed.

        .. versionchanged:: 2.9
            This now also includes the correct :attr:`~PartialEmoji.animated` value.
    """

    __slots__ = ("message_id", "channel_id", "guild_id", "emoji")

    def __init__(self, data: MessageReactionRemoveEmojiEvent, emoji: PartialEmoji) -> None:
        self.emoji: PartialEmoji = emoji
        self.message_id: int = int(data["message_id"])
        self.channel_id: int = int(data["channel_id"])
        try:
            self.guild_id: Optional[int] = int(data["guild_id"])
        except KeyError:
            self.guild_id: Optional[int] = None


class RawIntegrationDeleteEvent(_RawReprMixin):
    """Represents the event payload for an :func:`on_raw_integration_delete` event.

    .. versionadded:: 2.0

    Attributes
    ----------
    integration_id: :class:`int`
        The ID of the integration that got deleted.
    application_id: Optional[:class:`int`]
        The ID of the bot/OAuth2 application for this deleted integration.
    guild_id: :class:`int`
        The guild ID where the integration deletion took place.
    """

    __slots__ = ("integration_id", "application_id", "guild_id")

    def __init__(self, data: IntegrationDeleteEvent) -> None:
        self.integration_id: int = int(data["id"])
        self.guild_id: int = int(data["guild_id"])
        try:
            self.application_id: Optional[int] = int(data["application_id"])
        except KeyError:
            self.application_id: Optional[int] = None


class RawGuildScheduledEventUserActionEvent(_RawReprMixin):
    """Represents the event payload for an :func:`on_raw_guild_scheduled_event_subscribe`
    and :func:`on_raw_guild_scheduled_event_unsubscribe` events.

    .. versionadded:: 2.3

    Attributes
    ----------
    event_id: :class:`int`
        The ID of the guild scheduled event that the user subscribed to or unsubscribed from.
    user_id: :class:`int`
        The ID of the user doing the action.
    guild_id: :class:`int`
        The guild ID where the guild scheduled event is located.
    """

    __slots__ = ("event_id", "user_id", "guild_id")

    def __init__(
        self, data: Union[GuildScheduledEventUserAddEvent, GuildScheduledEventUserRemoveEvent]
    ) -> None:
        self.event_id: int = int(data["guild_scheduled_event_id"])
        self.user_id: int = int(data["user_id"])
        self.guild_id: int = int(data["guild_id"])


class RawThreadDeleteEvent(_RawReprMixin):
    """Represents the payload for a :func:`on_raw_thread_delete` event.

    .. versionadded:: 2.5

    Attributes
    ----------
    thread_id: :class:`int`
        The ID of the thread that was deleted.
    guild_id: :class:`int`
        The ID of the guild the thread was deleted in.
    thread_type: :class:`ChannelType`
        The type of the deleted thread.
    parent_id: :class:`int`
        The ID of the channel the thread belonged to.
    thread: Optional[:class:`Thread`]
        The thread, if it could be found in the internal cache.
    """

    __slots__ = (
        "thread_id",
        "thread_type",
        "parent_id",
        "guild_id",
        "thread",
    )

    def __init__(self, data: ThreadDeleteEvent) -> None:
        self.thread_id: int = int(data["id"])
        self.thread_type: ThreadType = cast("ThreadType", try_enum(ChannelType, data["type"]))
        self.guild_id: int = int(data["guild_id"])
        self.parent_id: int = int(data["parent_id"])
        self.thread: Optional[Thread] = None


class RawThreadMemberRemoveEvent(_RawReprMixin):
    """Represents the event payload for an :func:`on_raw_thread_member_remove` event.

    .. versionadded:: 2.5

    Attributes
    ----------
    thread: :class:`Thread`
        The Thread that the member was removed from
    member_id: :class:`int`
        The ID of the removed member.
    cached_member: Optional[:class:`.ThreadMember`]
        The member, if they could be found in the internal cache.
    """

    __slots__ = (
        "thread",
        "member_id",
        "cached_member",
    )

    def __init__(self, thread: Thread, member_id: int) -> None:
        self.thread: Thread = thread
        self.member_id: int = member_id
        self.cached_member: Optional[ThreadMember] = None


class RawTypingEvent(_RawReprMixin):
    """Represents the event payload for an :func:`on_raw_typing` event.

    .. versionadded:: 2.3

    Attributes
    ----------
    user_id: :class:`int`
        The ID of the user who started typing.
    channel_id: :class:`int`
        The ID of the channel where the user started typing.
    guild_id: Optional[:class:`int`]
        The ID of the guild where the user started typing or ``None`` if it was in a DM.
    member: Optional[:class:`Member`]
        The member object of the user who started typing or ``None`` if it was in a DM.
    timestamp: :class:`datetime.datetime`
        The UTC datetime when the user started typing.

        .. versionchanged:: 2.9
            Changed from naive to aware datetime.
    """

    __slots__ = ("user_id", "channel_id", "guild_id", "member", "timestamp")

    def __init__(self, data: TypingStartEvent) -> None:
        self.user_id: int = int(data["user_id"])
        self.channel_id: int = int(data["channel_id"])
        self.member: Optional[Member] = None
        self.timestamp: datetime.datetime = datetime.datetime.fromtimestamp(
            data["timestamp"], tz=datetime.timezone.utc
        )
        try:
            self.guild_id: Optional[int] = int(data["guild_id"])
        except KeyError:
            self.guild_id: Optional[int] = None


class RawGuildMemberRemoveEvent(_RawReprMixin):
    """Represents the event payload for an :func:`on_raw_member_remove` event.

    .. versionadded:: 2.6

    Attributes
    ----------
    guild_id: :class:`int`
        The ID of the guild where the member was removed from.
    user: Union[:class:`User`, :class:`Member`]
        The user object of the member that was removed.
    """

    __slots__ = (
        "guild_id",
        "user",
    )

    def __init__(self, user: Union[User, Member], guild_id: int) -> None:
        self.user: Union[User, Member] = user
        self.guild_id: int = guild_id


class RawPresenceUpdateEvent(_RawReprMixin):
    """Represents the event payload for an :func:`on_raw_presence_update` event.

    .. versionadded:: 2.10

    Attributes
    ----------
    user_id: :class:`int`
        The ID of the user whose presence was updated.
    guild_id: :class:`int`
        The ID of the guild where the user's presence changed.
    data: :class:`dict`
        The raw data given by the :ddocs:`gateway <topics/gateway-events#presence-update>`.
    """

    __slots__ = (
        "user_id",
        "guild_id",
        "data",
    )

    def __init__(self, data: PresenceUpdateEvent) -> None:
        self.user_id: int = int(data["user"]["id"])
        self.guild_id: int = int(data["guild_id"])
        self.data: PresenceUpdateEvent = data


class RawVoiceChannelEffectEvent(_RawReprMixin):
    """Represents the event payload for an :func:`on_raw_voice_channel_effect` event.

    .. versionadded:: 2.10

    Attributes
    ----------
    channel_id: :class:`int`
        The ID of the channel where the effect was sent.
    guild_id: :class:`int`
        The ID of the guild where the effect was sent.
    user_id: :class:`int`
        The ID of the user who sent the effect.
    effect: :class:`VoiceChannelEffect`
        The effect that was sent.
    cached_member: Optional[:class:`Member`]
        The member who sent the effect, if they could be found in the internal cache.
    """

    __slots__ = (
        "channel_id",
        "guild_id",
        "user_id",
        "effect",
        "cached_member",
    )

    def __init__(self, data: VoiceChannelEffectSendEvent, effect: VoiceChannelEffect) -> None:
        self.channel_id: int = int(data["channel_id"])
        self.guild_id: int = int(data["guild_id"])
        self.user_id: int = int(data["user_id"])
        self.effect: VoiceChannelEffect = effect

        self.cached_member: Optional[Member] = None



# SPDX-License-Identifier: MIT

"""disnake.ui.select
~~~~~~~~~~~~~~~~~~

Select Menu UI Kit Types

:copyright: (c) 2021-present Disnake Development
:license: MIT, see LICENSE for more details.

"""
from . import base, channel, mentionable, role, string, user
from .base import *
from .channel import *
from .mentionable import *
from .role import *
from .string import *
from .user import *

__all__ = []
__all__.extend(base.__all__)
__all__.extend(channel.__all__)
__all__.extend(mentionable.__all__)
__all__.extend(role.__all__)
__all__.extend(string.__all__)
__all__.extend(user.__all__)



# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import Any, List, Literal, Optional, Sequence, Tuple, TypedDict, Union

from typing_extensions import NotRequired

from .activity import PartialPresenceUpdate, PresenceData, SendableActivity
from .appinfo import PartialAppInfo, PartialGatewayAppInfo
from .audit_log import AuditLogEntry
from .automod import AutoModAction, AutoModRule, AutoModTriggerType
from .channel import Channel, GuildChannel, StageInstance
from .emoji import Emoji, PartialEmoji
from .entitlement import Entitlement
from .guild import Guild, UnavailableGuild
from .guild_scheduled_event import GuildScheduledEvent
from .integration import BaseIntegration
from .interactions import BaseInteraction, GuildApplicationCommandPermissions
from .invite import InviteTargetType, InviteType
from .member import MemberWithUser
from .message import Message
from .role import Role
from .snowflake import Snowflake, SnowflakeList
from .soundboard import GuildSoundboardSound
from .sticker import GuildSticker
from .threads import Thread, ThreadMember, ThreadMemberWithPresence, ThreadType
from .user import AvatarDecorationData, User
from .voice import GuildVoiceState, SupportedModes, VoiceChannelEffect


class SessionStartLimit(TypedDict):
    total: int
    remaining: int
    reset_after: int
    max_concurrency: int


class Gateway(TypedDict):
    url: str


class GatewayBot(Gateway):
    shards: int
    session_start_limit: SessionStartLimit


#####
# Websocket payloads (receive)
#####


class GatewayPayload(TypedDict):
    op: Literal[0, 1, 7, 9, 10, 11]
    d: Any  # event data
    s: Optional[int]  # sequence number
    t: Optional[str]  # event name


#####
# Websocket payloads (send)
#####

# opcode 1


class HeartbeatCommand(TypedDict):
    op: Literal[1, 3]  # normal ws and voice ws have different heartbeat opcodes
    d: Optional[int]


# opcode 2


class IdentifyProperties(TypedDict):
    os: str
    browser: str
    device: str


class IdentifyData(TypedDict):
    token: str
    properties: IdentifyProperties
    compress: NotRequired[bool]
    large_threshold: NotRequired[int]
    shard: NotRequired[Tuple[int, int]]
    presence: NotRequired[PresenceUpdateData]
    intents: int


class IdentifyCommand(TypedDict):
    op: Literal[2]
    d: IdentifyData


# opcode 3


class PresenceUpdateData(TypedDict):
    since: Optional[int]
    activities: Sequence[SendableActivity]
    status: str
    afk: bool


class PresenceUpdateCommand(TypedDict):
    op: Literal[3]
    d: PresenceUpdateData


# opcode 4


class VoiceStateData(TypedDict):
    guild_id: Snowflake
    channel_id: Optional[Snowflake]
    self_mute: bool
    self_deaf: bool


class VoiceStateCommand(TypedDict):
    op: Literal[4]
    d: VoiceStateData


# opcode 6


class ResumeData(TypedDict):
    token: str
    session_id: str
    seq: int


class ResumeCommand(TypedDict):
    op: Literal[6]
    d: ResumeData


# opcode 8


class RequestMembersData(TypedDict):
    guild_id: Snowflake
    query: NotRequired[str]
    limit: int
    presences: NotRequired[bool]
    user_ids: NotRequired[Union[Snowflake, SnowflakeList]]
    nonce: NotRequired[str]


class RequestMembersCommand(TypedDict):
    op: Literal[8]
    d: RequestMembersData


#####
# Voice payloads (receive)
#####


class VoicePayload(TypedDict):
    op: Literal[2, 4, 6, 8, 9]
    d: Any


class VoiceReadyPayload(TypedDict):
    ssrc: int
    ip: str
    port: int
    modes: List[str]


class VoiceSessionDescriptionPayload(TypedDict):
    mode: SupportedModes
    secret_key: List[int]


#####
# Voice payloads (send)
#####

# voice opcode 0


class VoiceIdentifyData(TypedDict):
    server_id: str
    user_id: str
    session_id: str
    token: str


class VoiceIdentifyCommand(TypedDict):
    op: Literal[0]
    d: VoiceIdentifyData


# voice opcode 1


class _VoiceSelectProtocolInnerData(TypedDict):
    address: str
    port: int
    mode: SupportedModes


class VoiceSelectProtocolData(TypedDict):
    protocol: Literal["udp"]
    data: _VoiceSelectProtocolInnerData


class VoiceSelectProtocolCommand(TypedDict):
    op: Literal[1]
    d: VoiceSelectProtocolData


# voice opcode 5


class VoiceSpeakingData(TypedDict):
    speaking: int  # bitfield of 3 bits
    delay: int
    ssrc: int


class VoiceSpeakingCommand(TypedDict):
    op: Literal[5]
    d: VoiceSpeakingData


# voice opcode 7


class VoiceResumeData(TypedDict):
    server_id: str
    session_id: str
    token: str


class VoiceResumeCommand(TypedDict):
    op: Literal[7]
    d: VoiceResumeData


#####
# Gateway events
#####


# https://discord.com/developers/docs/topics/gateway-events#ready
class ReadyEvent(TypedDict):
    v: int
    user: User
    guilds: List[UnavailableGuild]
    session_id: str
    resume_gateway_url: str
    shard: NotRequired[Tuple[int, int]]
    application: PartialGatewayAppInfo


# https://discord.com/developers/docs/topics/gateway-events#resumed
class ResumedEvent(TypedDict):
    ...


# https://discord.com/developers/docs/topics/gateway-events#application-command-permissions-update
ApplicationCommandPermissionsUpdateEvent = GuildApplicationCommandPermissions


# https://discord.com/developers/docs/topics/gateway-events#message-create
MessageCreateEvent = Message


# https://discord.com/developers/docs/topics/gateway-events#message-update
# This does not necessarily contain all message fields, but `id` and `channel_id` always exist
MessageUpdateEvent = Message


# https://discord.com/developers/docs/topics/gateway-events#message-delete
class MessageDeleteEvent(TypedDict):
    id: Snowflake
    channel_id: Snowflake
    guild_id: NotRequired[Snowflake]


# https://discord.com/developers/docs/topics/gateway-events#message-delete-bulk
class MessageDeleteBulkEvent(TypedDict):
    ids: List[Snowflake]
    channel_id: Snowflake
    guild_id: NotRequired[Snowflake]


class _BaseReactionEvent(TypedDict):
    user_id: Snowflake
    channel_id: Snowflake
    message_id: Snowflake
    guild_id: NotRequired[Snowflake]
    emoji: PartialEmoji


# https://discord.com/developers/docs/topics/gateway-events#message-reaction-add
class MessageReactionAddEvent(_BaseReactionEvent):
    member: NotRequired[MemberWithUser]
    message_author_id: NotRequired[Snowflake]


# https://discord.com/developers/docs/topics/gateway-events#message-reaction-remove
class MessageReactionRemoveEvent(_BaseReactionEvent):
    ...


# https://discord.com/developers/docs/topics/gateway-events#message-reaction-remove-all
class MessageReactionRemoveAllEvent(TypedDict):
    channel_id: Snowflake
    message_id: Snowflake
    guild_id: NotRequired[Snowflake]


# https://discord.com/developers/docs/topics/gateway-events#message-reaction-remove-emoji
class MessageReactionRemoveEmojiEvent(TypedDict):
    channel_id: Snowflake
    guild_id: NotRequired[Snowflake]
    message_id: Snowflake
    emoji: PartialEmoji


# https://discord.com/developers/docs/topics/gateway-events#message-poll-vote-add
class PollVoteAddEvent(TypedDict):
    channel_id: Snowflake
    guild_id: NotRequired[Snowflake]
    message_id: Snowflake
    user_id: Snowflake
    answer_id: int


# https://discord.com/developers/docs/topics/gateway-events#message-poll-vote-remove
class PollVoteRemoveEvent(TypedDict):
    channel_id: Snowflake
    guild_id: NotRequired[Snowflake]
    message_id: Snowflake
    user_id: Snowflake
    answer_id: int


# https://discord.com/developers/docs/topics/gateway-events#interaction-create
InteractionCreateEvent = BaseInteraction


# https://discord.com/developers/docs/topics/gateway-events#presence-update
PresenceUpdateEvent = PartialPresenceUpdate


# https://discord.com/developers/docs/topics/gateway-events#user-update
UserUpdateEvent = User


# https://discord.com/developers/docs/topics/gateway-events#invite-create
class InviteCreateEvent(TypedDict):
    channel_id: Snowflake
    code: str
    created_at: str
    guild_id: NotRequired[Snowflake]
    inviter: NotRequired[User]
    max_age: int
    max_uses: int
    target_type: NotRequired[InviteTargetType]
    target_user: NotRequired[User]
    target_application: NotRequired[PartialAppInfo]
    temporary: bool
    type: InviteType
    uses: int  # always 0


# https://discord.com/developers/docs/topics/gateway-events#invite-delete
class InviteDeleteEvent(TypedDict):
    channel_id: Snowflake
    guild_id: NotRequired[Snowflake]
    code: str


# https://discord.com/developers/docs/topics/gateway-events#channel-create
ChannelCreateEvent = GuildChannel


# https://discord.com/developers/docs/topics/gateway-events#channel-update
ChannelUpdateEvent = Channel


# https://discord.com/developers/docs/topics/gateway-events#channel-delete
ChannelDeleteEvent = Channel


# https://discord.com/developers/docs/topics/gateway-events#channel-pins-update
class ChannelPinsUpdateEvent(TypedDict):
    guild_id: NotRequired[Snowflake]
    channel_id: Snowflake
    last_pin_timestamp: NotRequired[Optional[str]]


# https://discord.com/developers/docs/topics/gateway-events#thread-create
class ThreadCreateEvent(Thread):
    newly_created: NotRequired[bool]


# https://discord.com/developers/docs/topics/gateway-events#thread-update
ThreadUpdateEvent = Thread


# https://discord.com/developers/docs/topics/gateway-events#thread-delete
class ThreadDeleteEvent(TypedDict):
    id: Snowflake
    guild_id: Snowflake
    parent_id: Snowflake
    type: ThreadType


# https://discord.com/developers/docs/topics/gateway-events#thread-list-sync
class ThreadListSyncEvent(TypedDict):
    guild_id: Snowflake
    channel_ids: NotRequired[List[Snowflake]]
    threads: List[Thread]
    members: List[ThreadMember]


# https://discord.com/developers/docs/topics/gateway-events#thread-member-update
class ThreadMemberUpdateEvent(ThreadMember):
    guild_id: Snowflake


# https://discord.com/developers/docs/topics/gateway-events#thread-members-update
class ThreadMembersUpdateEvent(TypedDict):
    id: Snowflake
    guild_id: Snowflake
    member_count: int
    added_members: NotRequired[List[ThreadMemberWithPresence]]
    removed_member_ids: NotRequired[List[Snowflake]]


# https://discord.com/developers/docs/topics/gateway-events#guild-member-add
class GuildMemberAddEvent(MemberWithUser):
    guild_id: Snowflake


# https://discord.com/developers/docs/topics/gateway-events#guild-member-remove
class GuildMemberRemoveEvent(TypedDict):
    guild_id: Snowflake
    user: User


# https://discord.com/developers/docs/topics/gateway-events#guild-member-update
class GuildMemberUpdateEvent(TypedDict):
    guild_id: Snowflake
    roles: List[Snowflake]
    user: User
    nick: NotRequired[Optional[str]]
    avatar: Optional[str]
    banner: Optional[str]
    joined_at: Optional[str]
    premium_since: NotRequired[Optional[str]]
    deaf: NotRequired[bool]
    mute: NotRequired[bool]
    pending: NotRequired[bool]
    communication_disabled_until: NotRequired[Optional[str]]
    flags: int
    avatar_decoration_data: NotRequired[Optional[AvatarDecorationData]]


# https://discord.com/developers/docs/topics/gateway-events#guild-emojis-update
class GuildEmojisUpdateEvent(TypedDict):
    guild_id: Snowflake
    emojis: List[Emoji]


# https://discord.com/developers/docs/topics/gateway-events#guild-stickers-update
class GuildStickersUpdateEvent(TypedDict):
    guild_id: Snowflake
    stickers: List[GuildSticker]


# https://discord.com/developers/docs/topics/gateway-events#guild-create
GuildCreateEvent = Union[Guild, UnavailableGuild]


# https://discord.com/developers/docs/topics/gateway-events#guild-update
GuildUpdateEvent = Guild


# https://discord.com/developers/docs/topics/gateway-events#guild-delete
GuildDeleteEvent = UnavailableGuild


# https://discord.com/developers/docs/topics/gateway-events#guild-audit-log-entry-create
class AuditLogEntryCreate(AuditLogEntry):
    guild_id: Snowflake


class _GuildBanEvent(TypedDict):
    guild_id: Snowflake
    user: User


# https://discord.com/developers/docs/topics/gateway-events#guild-ban-add
GuildBanAddEvent = _GuildBanEvent


# https://discord.com/developers/docs/topics/gateway-events#guild-ban-remove
GuildBanRemoveEvent = _GuildBanEvent


# https://discord.com/developers/docs/topics/gateway-events#guild-role-create
class GuildRoleCreateEvent(TypedDict):
    guild_id: Snowflake
    role: Role


# https://discord.com/developers/docs/topics/gateway-events#guild-role-delete
class GuildRoleDeleteEvent(TypedDict):
    guild_id: Snowflake
    role_id: Snowflake


# https://discord.com/developers/docs/topics/gateway-events#guild-role-update
class GuildRoleUpdateEvent(TypedDict):
    guild_id: Snowflake
    role: Role


# https://discord.com/developers/docs/topics/gateway-events#guild-scheduled-event-create
GuildScheduledEventCreateEvent = GuildScheduledEvent


# https://discord.com/developers/docs/topics/gateway-events#guild-scheduled-event-update
GuildScheduledEventUpdateEvent = GuildScheduledEvent


# https://discord.com/developers/docs/topics/gateway-events#guild-scheduled-event-delete
GuildScheduledEventDeleteEvent = GuildScheduledEvent


class _GuildScheduledEventUserEvent(TypedDict):
    guild_scheduled_event_id: Snowflake
    user_id: Snowflake
    guild_id: Snowflake


# https://discord.com/developers/docs/topics/gateway-events#guild-scheduled-event-user-add
GuildScheduledEventUserAddEvent = _GuildScheduledEventUserEvent


# https://discord.com/developers/docs/topics/gateway-events#guild-scheduled-event-user-remove
GuildScheduledEventUserRemoveEvent = _GuildScheduledEventUserEvent


# https://discord.com/developers/docs/topics/gateway-events#guild-members-chunk
class GuildMembersChunkEvent(TypedDict):
    guild_id: Snowflake
    members: List[MemberWithUser]
    chunk_index: int
    chunk_count: int
    not_found: NotRequired[List[Snowflake]]
    presences: NotRequired[List[PresenceData]]
    nonce: NotRequired[str]


# https://discord.com/developers/docs/topics/gateway-events#guild-integrations-update
class GuildIntegrationsUpdateEvent(TypedDict):
    guild_id: Snowflake


# https://discord.com/developers/docs/topics/gateway-events#integration-create
class IntegrationCreateEvent(BaseIntegration):
    guild_id: Snowflake


# https://discord.com/developers/docs/topics/gateway-events#integration-update
class IntegrationUpdateEvent(BaseIntegration):
    guild_id: Snowflake


# https://discord.com/developers/docs/topics/gateway-events#integration-delete
class IntegrationDeleteEvent(TypedDict):
    id: Snowflake
    guild_id: Snowflake
    application_id: NotRequired[Snowflake]


# https://discord.com/developers/docs/topics/gateway-events#webhooks-update
class WebhooksUpdateEvent(TypedDict):
    guild_id: Snowflake
    channel_id: Snowflake


# https://discord.com/developers/docs/topics/gateway-events#stage-instance-create
StageInstanceCreateEvent = StageInstance


# https://discord.com/developers/docs/topics/gateway-events#stage-instance-update
StageInstanceUpdateEvent = StageInstance


# https://discord.com/developers/docs/topics/gateway-events#stage-instance-delete
StageInstanceDeleteEvent = StageInstance


# https://discord.com/developers/docs/topics/gateway-events#voice-state-update
# We assume that we'll only receive voice states for guilds
VoiceStateUpdateEvent = GuildVoiceState


# https://discord.com/developers/docs/topics/gateway-events#voice-server-update
class VoiceServerUpdateEvent(TypedDict):
    token: str
    guild_id: Snowflake
    endpoint: Optional[str]


# https://discord.com/developers/docs/topics/gateway-events#voice-channel-effect-send
class VoiceChannelEffectSendEvent(VoiceChannelEffect):
    channel_id: Snowflake
    guild_id: Snowflake
    user_id: Snowflake


# https://discord.com/developers/docs/topics/gateway-events#typing-start
class TypingStartEvent(TypedDict):
    channel_id: Snowflake
    guild_id: NotRequired[Snowflake]
    user_id: Snowflake
    timestamp: int
    member: NotRequired[MemberWithUser]


# https://discord.com/developers/docs/topics/gateway-events#auto-moderation-rule-create
AutoModerationRuleCreateEvent = AutoModRule


# https://discord.com/developers/docs/topics/gateway-events#auto-moderation-rule-update
AutoModerationRuleUpdateEvent = AutoModRule


# https://discord.com/developers/docs/topics/gateway-events#auto-moderation-rule-delete
AutoModerationRuleDeleteEvent = AutoModRule


# https://discord.com/developers/docs/topics/gateway-events#auto-moderation-action-execution
class AutoModerationActionExecutionEvent(TypedDict):
    guild_id: Snowflake
    action: AutoModAction
    rule_id: Snowflake
    rule_trigger_type: AutoModTriggerType
    user_id: Snowflake
    channel_id: NotRequired[Optional[Snowflake]]
    message_id: NotRequired[Optional[Snowflake]]
    alert_system_message_id: NotRequired[Optional[Snowflake]]
    content: NotRequired[str]
    matched_content: NotRequired[Optional[str]]
    matched_keyword: NotRequired[Optional[str]]


# https://discord.com/developers/docs/monetization/entitlements#new-entitlement
EntitlementCreate = Entitlement


# https://discord.com/developers/docs/monetization/entitlements#updated-entitlement
EntitlementUpdate = Entitlement


# https://discord.com/developers/docs/monetization/entitlements#deleted-entitlement
EntitlementDelete = Entitlement


# https://discord.com/developers/docs/topics/gateway-events#guild-soundboard-sound-create
GuildSoundboardSoundCreate = GuildSoundboardSound


# https://discord.com/developers/docs/topics/gateway-events#guild-soundboard-sound-update
GuildSoundboardSoundUpdate = GuildSoundboardSound


# https://discord.com/developers/docs/topics/gateway-events#guild-soundboard-sound-delete
class GuildSoundboardSoundDelete(TypedDict):
    guild_id: Snowflake
    sound_id: Snowflake


# https://discord.com/developers/docs/topics/gateway-events#guild-soundboard-sounds-update
class GuildSoundboardSoundsUpdate(TypedDict):
    guild_id: Snowflake
    soundboard_sounds: List[GuildSoundboardSound]



# SPDX-License-Identifier: MIT

from __future__ import annotations

import functools
import inspect
import re
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Dict,
    Generic,
    Iterable,
    List,
    Literal,
    Optional,
    Protocol,
    Tuple,
    Type,
    TypeVar,
    Union,
    runtime_checkable,
)

import disnake

from .context import AnyContext, Context
from .errors import (
    BadArgument,
    BadBoolArgument,
    BadColourArgument,
    BadInviteArgument,
    BadLiteralArgument,
    BadUnionArgument,
    ChannelNotFound,
    ChannelNotReadable,
    CommandError,
    ConversionError,
    EmojiNotFound,
    GuildNotFound,
    GuildScheduledEventNotFound,
    GuildSoundboardSoundNotFound,
    GuildStickerNotFound,
    MemberNotFound,
    MessageNotFound,
    NoPrivateMessage,
    ObjectNotFound,
    PartialEmojiConversionFailure,
    RoleNotFound,
    ThreadNotFound,
    UserNotFound,
)

if TYPE_CHECKING:
    from disnake.abc import MessageableChannel


# TODO: USE ACTUAL FUNCTIONS INSTEAD OF USELESS CLASSES

__all__ = (
    "Converter",
    "IDConverter",
    "ObjectConverter",
    "MemberConverter",
    "UserConverter",
    "PartialMessageConverter",
    "MessageConverter",
    "GuildChannelConverter",
    "TextChannelConverter",
    "VoiceChannelConverter",
    "StageChannelConverter",
    "CategoryChannelConverter",
    "ForumChannelConverter",
    "MediaChannelConverter",
    "ThreadConverter",
    "ColourConverter",
    "ColorConverter",
    "RoleConverter",
    "GameConverter",
    "InviteConverter",
    "GuildConverter",
    "EmojiConverter",
    "PartialEmojiConverter",
    "GuildStickerConverter",
    "GuildSoundboardSoundConverter",
    "PermissionsConverter",
    "GuildScheduledEventConverter",
    "clean_content",
    "Greedy",
    "run_converters",
)


_utils_get = disnake.utils.get
T = TypeVar("T")
T_co = TypeVar("T_co", covariant=True)
CT = TypeVar("CT", bound=disnake.abc.GuildChannel)
TT = TypeVar("TT", bound=disnake.Thread)


def _get_from_guilds(
    client: disnake.Client, func: Callable[[disnake.Guild], Optional[T]]
) -> Optional[T]:
    for guild in client.guilds:
        if result := func(guild):
            return result
    return None


@runtime_checkable
class Converter(Protocol[T_co]):
    """The base class of custom converters that require the :class:`.Context`
    or :class:`.ApplicationCommandInteraction` to be passed to be useful.

    This allows you to implement converters that function similar to the
    special cased ``disnake`` classes.

    Classes that derive from this should override the :meth:`~.Converter.convert`
    method to do its conversion logic. This method must be a :ref:`coroutine <coroutine>`.
    """

    async def convert(self, ctx: AnyContext, argument: str) -> T_co:
        """|coro|

        The method to override to do conversion logic.

        If an error is found while converting, it is recommended to
        raise a :exc:`.CommandError` derived exception as it will
        properly propagate to the error handlers.

        Parameters
        ----------
        ctx: Union[:class:`.Context`, :class:`.ApplicationCommandInteraction`]
            The invocation context that the argument is being used in.
        argument: :class:`str`
            The argument that is being converted.

        Raises
        ------
        CommandError
            A generic exception occurred when converting the argument.
        BadArgument
            The converter failed to convert the argument.
        """
        raise NotImplementedError("Derived classes need to implement this.")


_ID_REGEX = re.compile(r"([0-9]{17,19})$")


class IDConverter(Converter[T_co]):
    @staticmethod
    def _get_id_match(argument: str) -> Optional[re.Match[str]]:
        return _ID_REGEX.match(argument)


class ObjectConverter(IDConverter[disnake.Object]):
    """Converts to a :class:`~disnake.Object`.

    The argument must follow the valid ID or mention formats (e.g. `<@80088516616269824>`).

    .. versionadded:: 2.0

    The lookup strategy is as follows (in order):

    1. Lookup by ID.
    2. Lookup by member, role, or channel mention.
    """

    async def convert(self, ctx: AnyContext, argument: str) -> disnake.Object:
        match = self._get_id_match(argument) or re.match(
            r"<(?:@(?:!|&)?|#)([0-9]{17,19})>$", argument
        )

        if match is None:
            raise ObjectNotFound(argument)

        result = int(match.group(1))

        return disnake.Object(id=result)


class MemberConverter(IDConverter[disnake.Member]):
    """Converts to a :class:`~disnake.Member`.

    All lookups are via the local guild. If in a DM context, then the lookup
    is done by the global cache.

    The lookup strategy is as follows (in order):

    1. Lookup by ID
    2. Lookup by mention
    3. Lookup by username#discrim
    4. Lookup by username#0
    5. Lookup by nickname
    6. Lookup by global name
    7. Lookup by username

    The name resolution order matches the one used by :meth:`.Guild.get_member_named`.

    .. versionchanged:: 1.5
        Raise :exc:`.MemberNotFound` instead of generic :exc:`.BadArgument`

    .. versionchanged:: 1.5.1
        This converter now lazily fetches members from the gateway and HTTP APIs,
        optionally caching the result if :attr:`.MemberCacheFlags.joined` is enabled.

    .. versionchanged:: 2.9
        Name resolution order changed from ``username > nick`` to
        ``nick > global_name > username`` to account for the username migration.
    """

    async def query_member_named(
        self, guild: disnake.Guild, argument: str
    ) -> Optional[disnake.Member]:
        cache = guild._state.member_cache_flags.joined

        username, _, discriminator = argument.rpartition("#")
        if username and (
            discriminator == "0" or (len(discriminator) == 4 and discriminator.isdecimal())
        ):
            # legacy behavior
            members = await guild.query_members(username, limit=100, cache=cache)
            return _utils_get(members, name=username, discriminator=discriminator)
        else:
            members = await guild.query_members(argument, limit=100, cache=cache)
            return disnake.utils.find(
                lambda m: m.nick == argument or m.global_name == argument or m.name == argument,
                members,
            )

    async def query_member_by_id(
        self, bot: disnake.Client, guild: disnake.Guild, user_id: int
    ) -> Optional[disnake.Member]:
        ws = bot._get_websocket(shard_id=guild.shard_id)
        cache = guild._state.member_cache_flags.joined
        if ws.is_ratelimited():
            # If we're being rate limited on the WS, then fall back to using the HTTP API
            # So we don't have to wait ~60 seconds for the query to finish
            try:
                member = await guild.fetch_member(user_id)
            except disnake.HTTPException:
                return None

            if cache:
                guild._add_member(member)
            return member

        # If we're not being rate limited then we can use the websocket to actually query
        members = await guild.query_members(limit=1, user_ids=[user_id], cache=cache)
        if not members:
            return None
        return members[0]

    async def convert(self, ctx: AnyContext, argument: str) -> disnake.Member:
        bot: disnake.Client = ctx.bot
        match = self._get_id_match(argument) or re.match(r"<@!?([0-9]{17,19})>$", argument)
        guild = ctx.guild
        result: Optional[disnake.Member] = None
        user_id: Optional[int] = None

        if match is None:
            # not a mention...
            if guild:
                result = guild.get_member_named(argument)
            else:
                result = _get_from_guilds(bot, lambda g: g.get_member_named(argument))
        else:
            user_id = int(match.group(1))
            if guild:
                mentions: Iterable[disnake.Member]
                if isinstance(ctx, Context):
                    mentions = (
                        user for user in ctx.message.mentions if isinstance(user, disnake.Member)
                    )
                else:
                    mentions = []
                result = guild.get_member(user_id) or _utils_get(mentions, id=user_id)
            else:
                result = _get_from_guilds(bot, lambda g: g.get_member(user_id))

        if result is None:
            if guild is None:
                raise MemberNotFound(argument)

            if user_id is not None:
                result = await self.query_member_by_id(bot, guild, user_id)
            else:
                result = await self.query_member_named(guild, argument)

            if not result:
                raise MemberNotFound(argument)

        return result


class UserConverter(IDConverter[disnake.User]):
    """Converts to a :class:`~disnake.User`.

    All lookups are via the global user cache.

    The lookup strategy is as follows (in order):

    1. Lookup by ID
    2. Lookup by mention
    3. Lookup by username#discrim
    4. Lookup by username#0
    5. Lookup by global name
    6. Lookup by username

    .. versionchanged:: 1.5
        Raise :exc:`.UserNotFound` instead of generic :exc:`.BadArgument`

    .. versionchanged:: 1.6
        This converter now lazily fetches users from the HTTP APIs if an ID is passed
        and it's not available in cache.

    .. versionchanged:: 2.9
        Now takes :attr:`~disnake.User.global_name` into account.
        No longer automatically removes ``"@"`` prefix from arguments.
    """

    async def convert(self, ctx: AnyContext, argument: str) -> disnake.User:
        match = self._get_id_match(argument) or re.match(r"<@!?([0-9]{17,19})>$", argument)
        state = ctx._state
        bot: disnake.Client = ctx.bot
        result: Optional[Union[disnake.User, disnake.Member]] = None

        if match is not None:
            user_id = int(match.group(1))

            mentions: Iterable[Union[disnake.User, disnake.Member]]
            if isinstance(ctx, Context):
                mentions = ctx.message.mentions
            else:
                mentions = []
            result = bot.get_user(user_id) or _utils_get(mentions, id=user_id)

            if result is None:
                try:
                    result = await bot.fetch_user(user_id)
                except disnake.HTTPException:
                    raise UserNotFound(argument) from None

            if isinstance(result, disnake.Member):
                return result._user
            return result

        username, _, discriminator = argument.rpartition("#")
        # n.b. there's no builtin method that only matches arabic digits, `isdecimal` is the closest one.
        # it really doesn't matter much, worst case is unnecessary computations
        if username and (
            discriminator == "0" or (len(discriminator) == 4 and discriminator.isdecimal())
        ):
            # legacy behavior
            result = _utils_get(state._users.values(), name=username, discriminator=discriminator)
            if result is not None:
                return result

        result = disnake.utils.find(
            lambda u: u.global_name == argument or u.name == argument,
            state._users.values(),
        )

        if result is None:
            raise UserNotFound(argument)

        return result


class PartialMessageConverter(Converter[disnake.PartialMessage]):
    """Converts to a :class:`~disnake.PartialMessage`.

    .. versionadded:: 1.7

    The creation strategy is as follows (in order):

    1. By "{channel ID}-{message ID}" (retrieved by shift-clicking on "Copy ID")
    2. By message ID (The message is assumed to be in the context channel.)
    3. By message URL
    """

    @staticmethod
    def _get_id_matches(ctx: AnyContext, argument: str) -> Tuple[Optional[int], int, int]:
        id_regex = re.compile(r"(?:(?P<channel_id>[0-9]{17,19})-)?(?P<message_id>[0-9]{17,19})$")
        link_regex = re.compile(
            r"https?://(?:(ptb|canary|www)\.)?discord(?:app)?\.com/channels/"
            r"(?P<guild_id>[0-9]{17,19}|@me)"
            r"/(?P<channel_id>[0-9]{17,19})/(?P<message_id>[0-9]{17,19})/?$"
        )
        match = id_regex.match(argument) or link_regex.match(argument)
        if not match:
            raise MessageNotFound(argument)
        data = match.groupdict()
        channel_id = disnake.utils._get_as_snowflake(data, "channel_id") or ctx.channel.id
        message_id = int(data["message_id"])
        guild_id_str: Optional[str] = data.get("guild_id")
        if guild_id_str is None:
            guild_id = ctx.guild and ctx.guild.id
        elif guild_id_str == "@me":
            guild_id = None
        else:
            guild_id = int(guild_id_str)
        return guild_id, message_id, channel_id

    @staticmethod
    def _resolve_channel(
        ctx: AnyContext, guild_id: Optional[int], channel_id: int
    ) -> Optional[MessageableChannel]:
        bot: disnake.Client = ctx.bot
        if guild_id is None:
            return bot.get_channel(channel_id) if channel_id else ctx.channel  # type: ignore

        guild = bot.get_guild(guild_id)
        if guild is not None:
            return guild._resolve_channel(channel_id)  # type: ignore
        return None

    async def convert(self, ctx: AnyContext, argument: str) -> disnake.PartialMessage:
        guild_id, message_id, channel_id = self._get_id_matches(ctx, argument)
        channel = self._resolve_channel(ctx, guild_id, channel_id)
        if not channel:
            raise ChannelNotFound(str(channel_id))
        return disnake.PartialMessage(channel=channel, id=message_id)


class MessageConverter(IDConverter[disnake.Message]):
    """Converts to a :class:`~disnake.Message`.

    .. versionadded:: 1.1

    The lookup strategy is as follows (in order):

    1. Lookup by "{channel ID}-{message ID}" (retrieved by shift-clicking on "Copy ID")
    2. Lookup by message ID (the message **must** be in the context channel)
    3. Lookup by message URL

    .. versionchanged:: 1.5
        Raise :exc:`.ChannelNotFound`, :exc:`.MessageNotFound` or :exc:`.ChannelNotReadable` instead of generic :exc:`.BadArgument`
    """

    async def convert(self, ctx: AnyContext, argument: str) -> disnake.Message:
        guild_id, message_id, channel_id = PartialMessageConverter._get_id_matches(ctx, argument)
        bot: disnake.Client = ctx.bot
        message = bot._connection._get_message(message_id)
        if message:
            return message
        channel = PartialMessageConverter._resolve_channel(ctx, guild_id, channel_id)
        if not channel:
            raise ChannelNotFound(str(channel_id))
        try:
            return await channel.fetch_message(message_id)
        except disnake.NotFound:
            raise MessageNotFound(argument) from None
        except disnake.Forbidden:
            raise ChannelNotReadable(channel) from None  # type: ignore


class GuildChannelConverter(IDConverter[disnake.abc.GuildChannel]):
    """Converts to a :class:`.abc.GuildChannel`.

    All lookups are via the local guild. If in a DM context, then the lookup
    is done by the global cache.

    The lookup strategy is as follows (in order):

    1. Lookup by ID.
    2. Lookup by mention.
    3. Lookup by name.

    .. versionadded:: 2.0
    """

    async def convert(self, ctx: AnyContext, argument: str) -> disnake.abc.GuildChannel:
        return self._resolve_channel(ctx, argument, "channels", disnake.abc.GuildChannel)

    @staticmethod
    def _resolve_channel(ctx: AnyContext, argument: str, attribute: str, type: Type[CT]) -> CT:
        bot: disnake.Client = ctx.bot

        match = IDConverter._get_id_match(argument) or re.match(r"<#([0-9]{17,19})>$", argument)
        result: Optional[disnake.abc.GuildChannel] = None
        guild = ctx.guild

        if match is None:
            # not a mention
            if guild:
                iterable: Iterable[CT] = getattr(guild, attribute)
                result = _utils_get(iterable, name=argument)
            else:
                result = disnake.utils.find(
                    lambda c: isinstance(c, type) and c.name == argument, bot.get_all_channels()
                )
        else:
            channel_id = int(match.group(1))
            if guild:
                result = guild.get_channel(channel_id)
            else:
                result = _get_from_guilds(bot, lambda g: g.get_channel(channel_id))

        if not isinstance(result, type):
            raise ChannelNotFound(argument)

        return result

    @staticmethod
    def _resolve_thread(ctx: AnyContext, argument: str, attribute: str, type: Type[TT]) -> TT:
        match = IDConverter._get_id_match(argument) or re.match(r"<#([0-9]{17,19})>$", argument)
        result: Optional[disnake.Thread] = None
        guild = ctx.guild

        if match is None:
            # not a mention
            if guild:
                iterable: Iterable[TT] = getattr(guild, attribute)
                result = _utils_get(iterable, name=argument)
        else:
            thread_id = int(match.group(1))
            if guild:
                result = guild.get_thread(thread_id)

        if not isinstance(result, type):
            raise ThreadNotFound(argument)

        return result


class TextChannelConverter(IDConverter[disnake.TextChannel]):
    """Converts to a :class:`~disnake.TextChannel`.

    All lookups are via the local guild. If in a DM context, then the lookup
    is done by the global cache.

    The lookup strategy is as follows (in order):

    1. Lookup by ID.
    2. Lookup by mention.
    3. Lookup by name

    .. versionchanged:: 1.5
        Raise :exc:`.ChannelNotFound` instead of generic :exc:`.BadArgument`
    """

    async def convert(self, ctx: AnyContext, argument: str) -> disnake.TextChannel:
        return GuildChannelConverter._resolve_channel(
            ctx, argument, "text_channels", disnake.TextChannel
        )


class VoiceChannelConverter(IDConverter[disnake.VoiceChannel]):
    """Converts to a :class:`~disnake.VoiceChannel`.

    All lookups are via the local guild. If in a DM context, then the lookup
    is done by the global cache.

    The lookup strategy is as follows (in order):

    1. Lookup by ID.
    2. Lookup by mention.
    3. Lookup by name

    .. versionchanged:: 1.5
        Raise :exc:`.ChannelNotFound` instead of generic :exc:`.BadArgument`
    """

    async def convert(self, ctx: AnyContext, argument: str) -> disnake.VoiceChannel:
        return GuildChannelConverter._resolve_channel(
            ctx, argument, "voice_channels", disnake.VoiceChannel
        )


class StageChannelConverter(IDConverter[disnake.StageChannel]):
    """Converts to a :class:`~disnake.StageChannel`.

    .. versionadded:: 1.7

    All lookups are via the local guild. If in a DM context, then the lookup
    is done by the global cache.

    The lookup strategy is as follows (in order):

    1. Lookup by ID.
    2. Lookup by mention.
    3. Lookup by name
    """

    async def convert(self, ctx: AnyContext, argument: str) -> disnake.StageChannel:
        return GuildChannelConverter._resolve_channel(
            ctx, argument, "stage_channels", disnake.StageChannel
        )


class CategoryChannelConverter(IDConverter[disnake.CategoryChannel]):
    """Converts to a :class:`~disnake.CategoryChannel`.

    All lookups are via the local guild. If in a DM context, then the lookup
    is done by the global cache.

    The lookup strategy is as follows (in order):

    1. Lookup by ID.
    2. Lookup by mention.
    3. Lookup by name

    .. versionchanged:: 1.5
        Raise :exc:`.ChannelNotFound` instead of generic :exc:`.BadArgument`
    """

    async def convert(self, ctx: AnyContext, argument: str) -> disnake.CategoryChannel:
        return GuildChannelConverter._resolve_channel(
            ctx, argument, "categories", disnake.CategoryChannel
        )


class ForumChannelConverter(IDConverter[disnake.ForumChannel]):
    """Converts to a :class:`~disnake.ForumChannel`.

    .. versionadded:: 2.5

    All lookups are via the local guild. If in a DM context, then the lookup
    is done by the global cache.

    The lookup strategy is as follows (in order):

    1. Lookup by ID.
    2. Lookup by mention.
    3. Lookup by name
    """

    async def convert(self, ctx: AnyContext, argument: str) -> disnake.ForumChannel:
        return GuildChannelConverter._resolve_channel(
            ctx, argument, "forum_channels", disnake.ForumChannel
        )


class MediaChannelConverter(IDConverter[disnake.MediaChannel]):
    """Converts to a :class:`~disnake.MediaChannel`.

    .. versionadded:: 2.10

    All lookups are via the local guild. If in a DM context, then the lookup
    is done by the global cache.

    The lookup strategy is as follows (in order):

    1. Lookup by ID.
    2. Lookup by mention.
    3. Lookup by name
    """

    async def convert(self, ctx: AnyContext, argument: str) -> disnake.MediaChannel:
        return GuildChannelConverter._resolve_channel(
            ctx, argument, "media_channels", disnake.MediaChannel
        )


class ThreadConverter(IDConverter[disnake.Thread]):
    """Coverts to a :class:`~disnake.Thread`.

    All lookups are via the local guild.

    The lookup strategy is as follows (in order):

    1. Lookup by ID.
    2. Lookup by mention.
    3. Lookup by name.

    .. versionadded:: 2.0
    """

    async def convert(self, ctx: AnyContext, argument: str) -> disnake.Thread:
        return GuildChannelConverter._resolve_thread(ctx, argument, "threads", disnake.Thread)


class ColourConverter(Converter[disnake.Colour]):
    """Converts to a :class:`~disnake.Colour`.

    .. versionchanged:: 1.5
        Add an alias named ColorConverter

    The following formats are accepted:

    - ``0x<hex>``
    - ``#<hex>``
    - ``0x#<hex>``
    - ``rgb(<number>, <number>, <number>)``
    - Any of the ``classmethod`` in :class:`~disnake.Colour`

        - The ``_`` in the name can be optionally replaced with spaces.

    Like CSS, ``<number>`` can be either 0-255 or 0-100% and ``<hex>`` can be
    either a 6 digit hex number or a 3 digit hex shortcut (e.g. #fff).

    .. versionchanged:: 1.5
        Raise :exc:`.BadColourArgument` instead of generic :exc:`.BadArgument`

    .. versionchanged:: 1.7
        Added support for ``rgb`` function and 3-digit hex shortcuts
    """

    RGB_REGEX = re.compile(
        r"rgb\s*\((?P<r>[0-9]{1,3}%?)\s*,\s*(?P<g>[0-9]{1,3}%?)\s*,\s*(?P<b>[0-9]{1,3}%?)\s*\)"
    )

    def parse_hex_number(self, argument: str) -> disnake.Color:
        arg = "".join(i * 2 for i in argument) if len(argument) == 3 else argument
        try:
            value = int(arg, base=16)
            if not (0 <= value <= 0xFFFFFF):
                raise BadColourArgument(argument)
        except ValueError:
            raise BadColourArgument(argument) from None
        else:
            return disnake.Color(value=value)

    def parse_rgb_number(self, argument: str, number: str) -> int:
        if number[-1] == "%":
            value = int(number[:-1])
            if not (0 <= value <= 100):
                raise BadColourArgument(argument)
            return round(255 * (value / 100))

        value = int(number)
        if not (0 <= value <= 255):
            raise BadColourArgument(argument)
        return value

    def parse_rgb(self, argument: str, *, regex: re.Pattern[str] = RGB_REGEX) -> disnake.Color:
        match = regex.match(argument)
        if match is None:
            raise BadColourArgument(argument)

        red = self.parse_rgb_number(argument, match.group("r"))
        green = self.parse_rgb_number(argument, match.group("g"))
        blue = self.parse_rgb_number(argument, match.group("b"))
        return disnake.Color.from_rgb(red, green, blue)

    async def convert(self, ctx: AnyContext, argument: str) -> disnake.Color:
        if argument[0] == "#":
            return self.parse_hex_number(argument[1:])

        if argument[0:2] == "0x":
            rest = argument[2:]
            # Legacy backwards compatible syntax
            if rest.startswith("#"):
                return self.parse_hex_number(rest[1:])
            return self.parse_hex_number(rest)

        arg = argument.lower()
        if arg[0:3] == "rgb":
            return self.parse_rgb(arg)

        arg = arg.replace(" ", "_")
        method = getattr(disnake.Colour, arg, None)
        if arg.startswith("from_") or method is None or not inspect.ismethod(method):
            raise BadColourArgument(arg)
        return method()


ColorConverter = ColourConverter


class RoleConverter(IDConverter[disnake.Role]):
    """Converts to a :class:`~disnake.Role`.

    All lookups are via the local guild. If in a DM context, the converter raises
    :exc:`.NoPrivateMessage` exception.

    The lookup strategy is as follows (in order):

    1. Lookup by ID.
    2. Lookup by mention.
    3. Lookup by name

    .. versionchanged:: 1.5
        Raise :exc:`.RoleNotFound` instead of generic :exc:`.BadArgument`
    """

    async def convert(self, ctx: AnyContext, argument: str) -> disnake.Role:
        guild = ctx.guild
        if not guild:
            raise NoPrivateMessage

        match = self._get_id_match(argument) or re.match(r"<@&([0-9]{17,19})>$", argument)
        if match:
            result = guild.get_role(int(match.group(1)))
        else:
            result = _utils_get(guild._roles.values(), name=argument)

        if result is None:
            raise RoleNotFound(argument)
        return result


class GameConverter(Converter[disnake.Game]):
    """Converts to :class:`~disnake.Game`."""

    async def convert(self, ctx: AnyContext, argument: str) -> disnake.Game:
        return disnake.Game(name=argument)


class InviteConverter(Converter[disnake.Invite]):
    """Converts to a :class:`~disnake.Invite`.

    This is done via an HTTP request using :meth:`.Bot.fetch_invite`.

    .. versionchanged:: 1.5
        Raise :exc:`.BadInviteArgument` instead of generic :exc:`.BadArgument`
    """

    async def convert(self, ctx: AnyContext, argument: str) -> disnake.Invite:
        try:
            return await ctx.bot.fetch_invite(argument)
        except Exception as exc:
            raise BadInviteArgument(argument) from exc


class GuildConverter(IDConverter[disnake.Guild]):
    """Converts to a :class:`~disnake.Guild`.

    The lookup strategy is as follows (in order):

    1. Lookup by ID.
    2. Lookup by name. (There is no disambiguation for Guilds with multiple matching names).

    .. versionadded:: 1.7
    """

    async def convert(self, ctx: AnyContext, argument: str) -> disnake.Guild:
        match = self._get_id_match(argument)
        bot: disnake.Client = ctx.bot
        result: Optional[disnake.Guild] = None

        if match is not None:
            guild_id = int(match.group(1))
            result = bot.get_guild(guild_id)

        if result is None:
            result = _utils_get(bot.guilds, name=argument)

            if result is None:
                raise GuildNotFound(argument)
        return result


class EmojiConverter(IDConverter[disnake.Emoji]):
    """Converts to a :class:`~disnake.Emoji`.

    All lookups are done for the local guild first, if available. If that lookup
    fails, then it checks the client's global cache.

    The lookup strategy is as follows (in order):

    1. Lookup by ID.
    2. Lookup by extracting ID from the emoji.
    3. Lookup by name

    .. versionchanged:: 1.5
        Raise :exc:`.EmojiNotFound` instead of generic :exc:`.BadArgument`
    """

    async def convert(self, ctx: AnyContext, argument: str) -> disnake.Emoji:
        match = self._get_id_match(argument) or re.match(
            r"<a?:[a-zA-Z0-9\_]{1,32}:([0-9]{17,19})>$", argument
        )
        result: Optional[disnake.Emoji] = None
        bot = ctx.bot
        guild = ctx.guild

        if match is None:
            # Try to get the emoji by name. Try local guild first.
            if guild:
                result = _utils_get(guild.emojis, name=argument)

            if result is None:
                result = _utils_get(bot.emojis, name=argument)
        else:
            # Try to look up emoji by id.
            result = bot.get_emoji(int(match.group(1)))

        if result is None:
            raise EmojiNotFound(argument)

        return result


class PartialEmojiConverter(Converter[disnake.PartialEmoji]):
    """Converts to a :class:`~disnake.PartialEmoji`.

    This is done by extracting the animated flag, name and ID from the emoji.

    .. versionchanged:: 1.5
        Raise :exc:`.PartialEmojiConversionFailure` instead of generic :exc:`.BadArgument`
    """

    async def convert(self, ctx: AnyContext, argument: str) -> disnake.PartialEmoji:
        match = re.match(r"<(a?):([a-zA-Z0-9\_]{1,32}):([0-9]{17,19})>$", argument)

        if match:
            emoji_animated = bool(match.group(1))
            emoji_name: str = match.group(2)
            emoji_id = int(match.group(3))

            return disnake.PartialEmoji.with_state(
                ctx.bot._connection, animated=emoji_animated, name=emoji_name, id=emoji_id
            )

        raise PartialEmojiConversionFailure(argument)


class GuildStickerConverter(IDConverter[disnake.GuildSticker]):
    """Converts to a :class:`~disnake.GuildSticker`.

    All lookups are done for the local guild first, if available. If that lookup
    fails, then it checks the client's global cache.

    The lookup strategy is as follows (in order):

    1. Lookup by ID
    2. Lookup by name

    .. versionadded:: 2.0
    """

    async def convert(self, ctx: AnyContext, argument: str) -> disnake.GuildSticker:
        match = self._get_id_match(argument)
        result = None
        bot: disnake.Client = ctx.bot
        guild = ctx.guild

        if match is None:
            # Try to get the sticker by name. Try local guild first.
            if guild:
                result = _utils_get(guild.stickers, name=argument)

            if result is None:
                result = _utils_get(bot.stickers, name=argument)
        else:
            # Try to look up sticker by id.
            result = bot.get_sticker(int(match.group(1)))

        if result is None:
            raise GuildStickerNotFound(argument)

        return result


class GuildSoundboardSoundConverter(IDConverter[disnake.GuildSoundboardSound]):
    """Converts to a :class:`~disnake.GuildSoundboardSound`.

    All lookups are done for the local guild first, if available. If that lookup
    fails, then it checks the client's global cache.

    The lookup strategy is as follows (in order):

    1. Lookup by ID
    2. Lookup by name

    .. versionadded:: 2.10
    """

    async def convert(self, ctx: AnyContext, argument: str) -> disnake.GuildSoundboardSound:
        match = self._get_id_match(argument)
        result = None
        bot: disnake.Client = ctx.bot
        guild = ctx.guild

        if match is None:
            # Try to get the sound by name. Try local guild first.
            if guild:
                result = _utils_get(guild.soundboard_sounds, name=argument)

            if result is None:
                result = _utils_get(bot.soundboard_sounds, name=argument)
        else:
            # Try to look up sound by id.
            result = bot.get_soundboard_sound(int(match.group(1)))

        if result is None:
            raise GuildSoundboardSoundNotFound(argument)

        return result


class PermissionsConverter(Converter[disnake.Permissions]):
    """Converts to a :class:`~disnake.Permissions`.

    Accepts an integer or a string of space-separated permission names (or just a single one) as input.

    .. versionadded:: 2.3
    """

    async def convert(self, ctx: AnyContext, argument: str) -> disnake.Permissions:
        # try the permission bit value
        try:
            value = int(argument)
        except ValueError:
            pass
        else:
            return disnake.Permissions(value)

        argument = argument.replace("server", "guild")

        # try multiple attributes, then a single one
        perms: List[disnake.Permissions] = []
        for name in argument.split():
            attr = getattr(disnake.Permissions, name, None)
            if attr is None:
                break

            if callable(attr):
                perms.append(attr())
            else:
                perms.append(disnake.Permissions(**{name: True}))
        else:
            return functools.reduce(lambda a, b: disnake.Permissions(a.value | b.value), perms)

        name = argument.replace(" ", "_")

        attr = getattr(disnake.Permissions, name, None)
        if attr is None:
            raise BadArgument(f"Invalid Permissions: {name!r}")

        if callable(attr):
            return attr()
        else:
            return disnake.Permissions(**{name: True})


class GuildScheduledEventConverter(IDConverter[disnake.GuildScheduledEvent]):
    """Converts to a :class:`~disnake.GuildScheduledEvent`.

    The lookup strategy is as follows (in order):

    1. Lookup by ID (in current guild)
    2. Lookup as event URL
    3. Lookup by name (in current guild; there is no disambiguation for scheduled events with multiple matching names)

    .. versionadded:: 2.5
    """

    async def convert(self, ctx: AnyContext, argument: str) -> disnake.GuildScheduledEvent:
        event_regex = re.compile(
            r"https?://(?:(?:ptb|canary|www)\.)?discord(?:app)?\.com/events/"
            r"([0-9]{17,19})/([0-9]{17,19})/?$"
        )
        bot: disnake.Client = ctx.bot
        result: Optional[disnake.GuildScheduledEvent] = None
        guild = ctx.guild

        # 1.
        if guild and (match := self._get_id_match(argument)):
            result = guild.get_scheduled_event(int(match.group(1)))

        # 2.
        if not result and (match := event_regex.match(argument)):
            event_guild = bot.get_guild(int(match.group(1)))
            if event_guild:
                result = event_guild.get_scheduled_event(int(match.group(2)))

        # 3.
        if not result and guild:
            result = _utils_get(guild.scheduled_events, name=argument)

        if not result:
            raise GuildScheduledEventNotFound(argument)
        return result


class clean_content(Converter[str]):
    """Converts the argument to mention scrubbed version of
    said content.

    This behaves similarly to :attr:`~disnake.Message.clean_content`.

    Attributes
    ----------
    fix_channel_mentions: :class:`bool`
        Whether to clean channel mentions.
    use_nicknames: :class:`bool`
        Whether to use :attr:`nicknames <.Member.nick>` and
        :attr:`global names <.Member.global_name>` when transforming mentions.
    escape_markdown: :class:`bool`
        Whether to also escape special markdown characters.
    remove_markdown: :class:`bool`
        Whether to also remove special markdown characters. This option is not supported with ``escape_markdown``

        .. versionadded:: 1.7
    """

    def __init__(
        self,
        *,
        fix_channel_mentions: bool = False,
        use_nicknames: bool = True,
        escape_markdown: bool = False,
        remove_markdown: bool = False,
    ) -> None:
        self.fix_channel_mentions = fix_channel_mentions
        self.use_nicknames = use_nicknames
        self.escape_markdown = escape_markdown
        self.remove_markdown = remove_markdown

    async def convert(self, ctx: AnyContext, argument: str) -> str:
        msg = ctx.message if isinstance(ctx, Context) else None
        bot: disnake.Client = ctx.bot

        def resolve_user(id: int) -> str:
            m = (
                (msg and _utils_get(msg.mentions, id=id))
                or (ctx.guild and ctx.guild.get_member(id))
                or bot.get_user(id)
            )
            return f"@{m.display_name if self.use_nicknames else m.name}" if m else "@deleted-user"

        def resolve_role(id: int) -> str:
            if ctx.guild is None:
                return "@deleted-role"
            r = (msg and _utils_get(msg.role_mentions, id=id)) or ctx.guild.get_role(id)
            return f"@{r.name}" if r else "@deleted-role"

        def resolve_channel(id: int) -> str:
            if ctx.guild and self.fix_channel_mentions:
                c = ctx.guild.get_channel(id)
                return f"#{c.name}" if c else "#deleted-channel"

            return f"<#{id}>"

        transforms: Dict[str, Callable[[int], str]] = {
            "@": resolve_user,
            "@!": resolve_user,
            "#": resolve_channel,
            "@&": resolve_role,
        }

        def repl(match: re.Match) -> str:
            type = match[1]
            id = int(match[2])
            return transforms[type](id)

        result = re.sub(r"<(@[!&]?|#)([0-9]{17,19})>", repl, argument)
        if self.escape_markdown:
            result = disnake.utils.escape_markdown(result)
        elif self.remove_markdown:
            result = disnake.utils.remove_markdown(result)

        # Completely ensure no mentions escape:
        return disnake.utils.escape_mentions(result)


class Greedy(List[T]):
    """A special converter that greedily consumes arguments until it can't.
    As a consequence of this behaviour, most input errors are silently discarded,
    since it is used as an indicator of when to stop parsing.

    When a parser error is met the greedy converter stops converting, undoes the
    internal string parsing routine, and continues parsing regularly.

    For example, in the following code:

    .. code-block:: python3

        @commands.command()
        async def test(ctx, numbers: Greedy[int], reason: str):
            await ctx.send("numbers: {0}, reason: {1}".format(numbers, reason))

    An invocation of ``[p]test 1 2 3 4 5 6 hello`` would pass ``numbers`` with
    ``[1, 2, 3, 4, 5, 6]`` and ``reason`` with ``hello``.

    For more information, check :ref:`ext_commands_special_converters`.
    """

    __slots__ = ("converter",)

    def __init__(self, *, converter: T) -> None:
        self.converter = converter

    def __repr__(self) -> str:
        converter = getattr(self.converter, "__name__", repr(self.converter))
        return f"Greedy[{converter}]"

    def __class_getitem__(cls, params: Union[Tuple[T], T]) -> Greedy[T]:
        if not isinstance(params, tuple):
            params = (params,)
        if len(params) != 1:
            raise TypeError("Greedy[...] only takes a single argument")
        converter = params[0]

        origin = getattr(converter, "__origin__", None)
        args = getattr(converter, "__args__", ())

        if not (callable(converter) or isinstance(converter, Converter) or origin is not None):
            raise TypeError("Greedy[...] expects a type or a Converter instance.")

        if converter in (str, type(None)) or origin is Greedy:
            raise TypeError(f"Greedy[{converter.__name__}] is invalid.")

        if origin is Union and type(None) in args:
            raise TypeError(f"Greedy[{converter!r}] is invalid.")

        return cls(converter=converter)


def _convert_to_bool(argument: str) -> bool:
    lowered = argument.lower()
    if lowered in ("yes", "y", "true", "t", "1", "enable", "on"):
        return True
    elif lowered in ("no", "n", "false", "f", "0", "disable", "off"):
        return False
    else:
        raise BadBoolArgument(lowered)


def get_converter(param: inspect.Parameter) -> Any:
    converter = param.annotation
    if converter is param.empty:
        if param.default is not param.empty:
            converter = str if param.default is None else type(param.default)
        else:
            converter = str
    return converter


_GenericAlias = type(List[Any])


def is_generic_type(tp: Any, *, _GenericAlias: Type = _GenericAlias) -> bool:
    return (isinstance(tp, type) and issubclass(tp, Generic)) or isinstance(tp, _GenericAlias)


CONVERTER_MAPPING: Dict[Type[Any], Type[Converter]] = {
    disnake.Object: ObjectConverter,
    disnake.Member: MemberConverter,
    disnake.User: UserConverter,
    disnake.Message: MessageConverter,
    disnake.PartialMessage: PartialMessageConverter,
    disnake.TextChannel: TextChannelConverter,
    disnake.Invite: InviteConverter,
    disnake.Guild: GuildConverter,
    disnake.Role: RoleConverter,
    disnake.Game: GameConverter,
    disnake.Colour: ColourConverter,
    disnake.VoiceChannel: VoiceChannelConverter,
    disnake.StageChannel: StageChannelConverter,
    disnake.Emoji: EmojiConverter,
    disnake.PartialEmoji: PartialEmojiConverter,
    disnake.CategoryChannel: CategoryChannelConverter,
    disnake.ForumChannel: ForumChannelConverter,
    disnake.MediaChannel: MediaChannelConverter,
    disnake.Thread: ThreadConverter,
    disnake.abc.GuildChannel: GuildChannelConverter,
    disnake.GuildSticker: GuildStickerConverter,
    disnake.GuildSoundboardSound: GuildSoundboardSoundConverter,
    disnake.Permissions: PermissionsConverter,
    disnake.GuildScheduledEvent: GuildScheduledEventConverter,
}


async def _actual_conversion(
    ctx: Context,
    converter: Union[Type[T], Type[Converter[T]], Converter[T], Callable[[str], T]],
    argument: str,
    param: inspect.Parameter,
) -> T:
    if converter is bool:
        return _convert_to_bool(argument)  # type: ignore

    if isinstance(converter, type):
        module = converter.__module__
        if module.startswith("disnake.") and not module.endswith("converter"):
            converter = CONVERTER_MAPPING.get(converter, converter)

    try:
        if isinstance(converter, type) and issubclass(converter, Converter):
            if inspect.ismethod(converter.convert):
                return await converter.convert(ctx, argument)
            else:
                return await converter().convert(ctx, argument)
        elif isinstance(converter, Converter):
            return await converter.convert(ctx, argument)  # type: ignore
    except CommandError:
        raise
    except Exception as exc:
        raise ConversionError(converter, exc) from exc

    try:
        return converter(argument)  # type: ignore
    except CommandError:
        raise
    except Exception as exc:
        try:
            name = converter.__name__
        except AttributeError:
            name = converter.__class__.__name__

        raise BadArgument(f'Converting to "{name}" failed for parameter "{param.name}".') from exc


async def run_converters(ctx: Context, converter, argument: str, param: inspect.Parameter):
    """|coro|

    Runs converters for a given converter, argument, and parameter.

    This function does the same work that the library does under the hood.

    .. versionadded:: 2.0

    Parameters
    ----------
    ctx: :class:`Context`
        The invocation context to run the converters under.
    converter: Any
        The converter to run, this corresponds to the annotation in the function.
    argument: :class:`str`
        The argument to convert to.
    param: :class:`inspect.Parameter`
        The parameter being converted. This is mainly for error reporting.

    Raises
    ------
    CommandError
        The converter failed to convert.

    Returns
    -------
    Any
        The resulting conversion.
    """
    origin = getattr(converter, "__origin__", None)

    if origin is Union:
        errors = []
        _NoneType = type(None)
        union_args = converter.__args__
        for conv in union_args:
            # if we got to this part in the code, then the previous conversions have failed
            # so we should just undo the view, return the default, and allow parsing to continue
            # with the other parameters
            if conv is _NoneType and param.kind != param.VAR_POSITIONAL:
                ctx.view.undo()
                return None if param.default is param.empty else param.default

            try:
                value = await run_converters(ctx, conv, argument, param)
            except CommandError as exc:
                errors.append(exc)
            else:
                return value

        # if we're here, then we failed all the converters
        raise BadUnionArgument(param, union_args, errors)

    if origin is Literal:
        errors = []
        conversions = {}
        literal_args = converter.__args__
        for literal in literal_args:
            literal_type = type(literal)
            try:
                value = conversions[literal_type]
            except KeyError:
                try:
                    value = await _actual_conversion(ctx, literal_type, argument, param)
                except CommandError as exc:
                    errors.append(exc)
                    conversions[literal_type] = object()
                    continue
                else:
                    conversions[literal_type] = value

            if value == literal:
                return value

        # if we're here, then we failed to match all the literals
        raise BadLiteralArgument(param, literal_args, errors)

    # This must be the last if-clause in the chain of origin checking
    # Nearly every type is a generic type within the typing library
    # So care must be taken to make sure a more specialised origin handle
    # isn't overwritten by the widest if clause
    if origin is not None and is_generic_type(converter):
        converter = origin

    return await _actual_conversion(ctx, converter, argument, param)



# SPDX-License-Identifier: MIT

from __future__ import annotations

import logging
import os
import warnings
from abc import ABC, abstractmethod
from collections import defaultdict
from pathlib import Path
from typing import (
    TYPE_CHECKING,
    Any,
    DefaultDict,
    Dict,
    Generic,
    Literal,
    Optional,
    Set,
    TypeVar,
    Union,
    overload,
)

from . import utils
from .custom_warnings import LocalizationWarning
from .enums import Locale
from .errors import LocalizationKeyError

if TYPE_CHECKING:
    from typing_extensions import Self

    LocalizedRequired = Union[str, "Localized[str]"]
    LocalizedOptional = Union[Optional[str], "Localized[Optional[str]]"]


__all__ = (
    "Localized",
    "Localised",
    "LocalizationValue",
    "LocalizationProtocol",
    "LocalizationStore",
)

MISSING = utils.MISSING

_log = logging.getLogger(__name__)


LocalizationsDict = Union[Dict[Locale, str], Dict[str, str]]
Localizations = Union[str, LocalizationsDict]

StringT = TypeVar("StringT", str, Optional[str], covariant=True)


# This is generic over `string`, as some localized strings can be optional, e.g. option descriptions.
# The basic idea for parameters is this:
#     abc: LocalizedRequired
#     xyz: LocalizedOptional = None
#
# With that, one may use `abc="somename"` and `abc=Localized("somename", ...)`,
# but not `abc=Localized(None, ...)`. All three work fine for `xyz` though.


class Localized(Generic[StringT]):
    """A container type used for localized parameters.

    Exactly one of ``key`` or ``data`` must be provided.

    There is an alias for this called ``Localised``.

    .. versionadded:: 2.5

    Parameters
    ----------
    string: Optional[:class:`str`]
        The default (non-localized) value of the string.
        Whether this is optional or not depends on the localized parameter type.
    key: :class:`str`
        A localization key used for lookups.
        Incompatible with ``data``.
    data: Union[Dict[:class:`.Locale`, :class:`str`], Dict[:class:`str`, :class:`str`]]
        A mapping of locales to localized values.
        Incompatible with ``key``.
    """

    __slots__ = ("string", "localizations")

    @overload
    def __init__(self: Localized[StringT], string: StringT, *, key: str) -> None:
        ...

    @overload
    def __init__(self: Localized[Optional[str]], *, key: str) -> None:
        ...

    @overload
    def __init__(
        self: Localized[StringT],
        string: StringT,
        *,
        data: Union[Optional[LocalizationsDict], LocalizationValue],
    ) -> None:
        ...

    @overload
    def __init__(
        self: Localized[Optional[str]],
        *,
        data: Union[Optional[LocalizationsDict], LocalizationValue],
    ) -> None:
        ...

    # note: `data` accepting `LocalizationValue` is intentionally undocumented,
    # as it's only meant to be used internally
    def __init__(
        self,
        string: StringT = None,
        *,
        key: str = MISSING,
        data: Union[Optional[LocalizationsDict], LocalizationValue] = MISSING,
    ) -> None:
        self.string: StringT = string

        if not (key is MISSING) ^ (data is MISSING):
            raise TypeError("Exactly one of `key` or `data` must be provided")
        if isinstance(data, LocalizationValue):
            self.localizations = data
        else:
            self.localizations = LocalizationValue(key if key is not MISSING else data)

    @overload
    @classmethod
    def _cast(cls, string: LocalizedOptional, required: Literal[False]) -> Localized[Optional[str]]:
        ...

    @overload
    @classmethod
    def _cast(cls, string: LocalizedRequired, required: Literal[True]) -> Localized[str]:
        ...

    @classmethod
    def _cast(cls, string: Union[Optional[str], Localized[Any]], required: bool) -> Localized[Any]:
        if not isinstance(string, Localized):
            string = cls(string, data=None)

        # enforce the `StringT` type at runtime
        if required and string.string is None:
            raise ValueError("`string` parameter must be provided")
        return string

    @overload
    def _upgrade(self, *, key: Optional[str]) -> Self:
        ...

    @overload
    def _upgrade(self, string: str, *, key: Optional[str] = None) -> Localized[str]:
        ...

    def _upgrade(
        self: Localized[Any], string: Optional[str] = None, *, key: Optional[str] = None
    ) -> Localized[Any]:
        # update key if provided and not already set
        self.localizations._upgrade(key)

        # Only overwrite if not already set (`Localized()` parameter value takes precedence over function names etc.)
        # Note: not checking whether `string` is an empty string, to keep generic typing correct
        if not self.string and string is not None:
            self.string = string

        # this is safe, see above
        return self


Localised = Localized


class LocalizationValue:
    """Container type for (pending) localization data.

    .. versionadded:: 2.5
    """

    __slots__ = ("_key", "_data")

    def __init__(self, localizations: Optional[Localizations]) -> None:
        self._key: Optional[str]
        self._data: Optional[Dict[str, str]]

        if localizations is None:
            # no localization
            self._key = None
            self._data = None
        elif isinstance(localizations, str):
            # got localization key
            self._key = localizations
            self._data = MISSING  # not localized yet
        elif isinstance(localizations, dict):
            # got localization data
            self._key = None
            self._data = {str(k): v for k, v in localizations.items()}
        else:
            raise TypeError(f"Invalid localizations type: {type(localizations).__name__}")

    def _upgrade(self, key: Optional[str]) -> None:
        if not key:
            return

        # if empty, use new key
        if self._key is None and self._data is None:
            self._key = key
            self._data = MISSING
            return

        # if key is the same, ignore
        if self._key == key:
            return

        # at this point, the keys don't match, which either means that they're different strings,
        # or that there is no existing `_key` but `_data` is set
        raise ValueError("Can't specify multiple localization keys or dicts")

    def _link(self, store: LocalizationProtocol) -> None:
        """Loads localizations from the specified store if this object has a key."""
        if self._key is not None:
            self._data = store.get(self._key)

    def _copy(self) -> LocalizationValue:
        cls = self.__class__
        ins = cls.__new__(cls)
        ins._key = self._key
        ins._data = self._data
        return ins

    @property
    def data(self) -> Optional[Dict[str, str]]:
        """Optional[Dict[:class:`str`, :class:`str`]]: A dict with a locale -> localization mapping, if available."""
        if self._data is MISSING:
            # This will happen when `_link(store)` hasn't been called yet, which *shouldn't* occur under normal circumstances.
            warnings.warn(
                f"Localization value ('{self._key}') was never linked to bot; this may be a library bug.",
                LocalizationWarning,
                stacklevel=2,
            )
            return None
        return self._data

    def __eq__(self, other) -> bool:
        # if both are pending, compare keys instead
        if self._data is MISSING and other._data is MISSING:
            return self._key == other._key

        d1 = self.data
        d2 = other.data
        # consider values equal if they're both falsy, or actually equal
        # (it doesn't matter if localizations are `None` or `{}`)
        return (not d1 and not d2) or d1 == d2


class LocalizationProtocol(ABC):
    """Manages a key-value mapping of localizations.

    This is an abstract class, a concrete implementation is provided as :class:`LocalizationStore`.

    .. versionadded:: 2.5
    """

    @abstractmethod
    def get(self, key: str) -> Optional[Dict[str, str]]:
        """Returns localizations for the specified key.

        Parameters
        ----------
        key: :class:`str`
            The lookup key.

        Raises
        ------
        LocalizationKeyError
            May be raised if no localizations for the provided key were found,
            depending on the implementation.

        Returns
        -------
        Optional[Dict[:class:`str`, :class:`str`]]
            The localizations for the provided key.
            May return ``None`` if no localizations could be found.
        """
        raise NotImplementedError

    # subtypes don't have to implement this
    def load(self, path: Union[str, os.PathLike]) -> None:
        """Adds localizations from the provided path.

        Parameters
        ----------
        path: Union[:class:`str`, :class:`os.PathLike`]
            The path to the file/directory to load.

        Raises
        ------
        RuntimeError
            The provided path is invalid or couldn't be loaded
        """
        raise NotImplementedError

    # subtypes don't have to implement this
    def reload(self) -> None:
        """Clears localizations and reloads all previously loaded sources again.
        If an exception occurs, the previous data gets restored and the exception is re-raised.
        """
        pass


class LocalizationStore(LocalizationProtocol):
    """Manages a key-value mapping of localizations using ``.json`` files.

    .. versionadded:: 2.5

    Attributes
    ----------
    strict: :class:`bool`
        Specifies whether :meth:`.get` raises an exception if localizations for a provided key couldn't be found.
    """

    def __init__(self, *, strict: bool) -> None:
        self.strict = strict

        self._loc: DefaultDict[str, Dict[str, str]] = defaultdict(dict)
        self._paths: Set[Path] = set()

    def get(self, key: str) -> Optional[Dict[str, str]]:
        """Returns localizations for the specified key.

        Parameters
        ----------
        key: :class:`str`
            The lookup key.

        Raises
        ------
        LocalizationKeyError
            No localizations for the provided key were found.
            Raised only if :attr:`strict` is enabled, returns ``None`` otherwise.

        Returns
        -------
        Optional[Dict[:class:`str`, :class:`str`]]
            The localizations for the provided key.
            Returns ``None`` if no localizations could be found and :attr:`strict` is disabled.
        """
        data = self._loc.get(key)
        if data is None and self.strict:
            raise LocalizationKeyError(key)
        return data

    def load(self, path: Union[str, os.PathLike]) -> None:
        """Adds localizations from the provided path to the store.
        If the path points to a file, the file gets loaded.
        If it's a directory, all ``.json`` files in that directory get loaded (non-recursive).

        Parameters
        ----------
        path: Union[:class:`str`, :class:`os.PathLike`]
            The path to the file/directory to load.

        Raises
        ------
        RuntimeError
            The provided path is invalid or couldn't be loaded
        """
        path = Path(path)

        if path.is_file():
            self._load_file(path)
        elif path.is_dir():
            for file in path.glob("*.json"):
                if not file.is_file():
                    continue
                self._load_file(file)
        else:
            raise RuntimeError(f"Path '{path}' does not exist or is not a directory/file")

        self._paths.add(path)

    def reload(self) -> None:
        """Clears localizations and reloads all previously loaded files/directories again.
        If an exception occurs, the previous data gets restored and the exception is re-raised.
        See :func:`~LocalizationStore.load` for possible raised exceptions.
        """
        old = self._loc
        try:
            self._loc = defaultdict(dict)
            for path in self._paths:
                self.load(path)
        except Exception:
            # restore in case of error
            self._loc = old
            raise

    def _load_file(self, path: Path) -> None:
        try:
            if path.suffix != ".json":
                raise ValueError("not a .json file")
            locale = path.stem

            if not (api_locale := utils.as_valid_locale(locale)):
                raise ValueError(f"invalid locale '{locale}'")
            locale = api_locale

            data = utils._from_json(path.read_text("utf-8"))
            self._load_dict(data, locale)
            _log.debug(f"Loaded localizations from '{path}'")
        except Exception as e:
            raise RuntimeError(f"Unable to load '{path}': {e}") from e

    def _load_dict(self, data: Dict[str, Optional[str]], locale: str) -> None:
        if not isinstance(data, dict) or not all(
            o is None or isinstance(o, str) for o in data.values()
        ):
            raise TypeError("data must be a flat dict with string/null values")
        for key, value in data.items():
            d = self._loc[key]  # always create dict, regardless of value
            if value is not None:
                d[locale] = value



# SPDX-License-Identifier: MIT

"""disnake.ext.commands
~~~~~~~~~~~~~~~~~~~~~

An extension module to facilitate creation of bot commands.

:copyright: (c) 2015-2021 Rapptz, 2021-present Disnake Development
:license: MIT, see LICENSE for more details.
"""

from .base_core import *
from .bot import *
from .cog import *
from .context import *
from .converter import *
from .cooldowns import *
from .core import *
from .ctx_menus_core import *
from .custom_warnings import *
from .errors import *
from .flag_converter import *
from .flags import *
from .help import *
from .params import *
from .slash_core import *



# SPDX-License-Identifier: MIT

from __future__ import annotations

from typing import TYPE_CHECKING, NamedTuple, Optional, Sequence

__all__ = ("BanEntry",)

if TYPE_CHECKING:
    from .abc import Snowflake
    from .user import User


class BanEntry(NamedTuple):
    reason: Optional[str]
    user: "User"


class BulkBanResult(NamedTuple):
    banned: Sequence[Snowflake]
    failed: Sequence[Snowflake]
